from unittest import TestCase

from mock import patch

from pyinfra.api import Config, State
from pyinfra.api.exceptions import PyinfraError

from ..util import make_inventory


class TestStateApi(TestCase):
    def test_require_pyinfra_requirement_ok(self):
        config = Config(REQUIRE_PYINFRA_VERSION='>=100')
        inventory = make_inventory()

        with patch('pyinfra.api.state.__version__', '100'):
            State(inventory, config)

    def test_require_pyinfra_requirement_too_low(self):
        config = Config(REQUIRE_PYINFRA_VERSION='>=100')
        inventory = make_inventory()

        with self.assertRaises(PyinfraError) as context:
            with patch('pyinfra.api.state.__version__', '99'):
                State(inventory, config)

        assert context.exception.args[0] == (
            'pyinfra version requirement not met (requires >=100, running 99)'
        )

    def test_require_pyinfra_min_version_ok(self):
        config = Config(MIN_PYINFRA_VERSION=100)
        inventory = make_inventory()

        with patch('pyinfra.api.state.__version__', '100'):
            State(inventory, config)

    def test_pyinfra_min_version_ignored_when_required_version_set(self):
        config = Config(
            REQUIRE_PYINFRA_VERSION='==100',
            MIN_PYINFRA_VERSION=1000,  # should be ignored
        )
        inventory = make_inventory()

        with patch('pyinfra.api.state.__version__', '100'):
            State(inventory, config)



from __future__ import unicode_literals

from pyinfra.api import FactBase

# Mapping for iptables code arguments to variable names
IPTABLES_ARGS = {
    '-A': 'chain',
    '-j': 'jump',

    # Boolean matches
    '-p': 'protocol',
    '-s': 'source',
    '-d': 'destination',
    '-i': 'in_interface',
    '-o': 'out_interface',

    # Logging
    '--log-prefix': 'log_prefix',

    # NAT exit rules
    '--to-destination': 'to_destination',
    '--to-source': 'to_source',
    '--to-ports': 'to_ports',
}


def parse_iptables_rule(line):
    '''
    Parse one iptables rule. Returns a dict where each iptables code argument
    is mapped to a name using IPTABLES_ARGS.
    '''

    bits = line.split()

    definition = {}

    key = None
    args = []
    not_arg = False

    def add_args():
        arg_string = ' '.join(args)

        if key in IPTABLES_ARGS:
            definition_key = (
                'not_{0}'.format(IPTABLES_ARGS[key])
                if not_arg
                else IPTABLES_ARGS[key]
            )
            definition[definition_key] = arg_string
        else:
            definition.setdefault('extras', []).extend((key, arg_string))

    for bit in bits:
        if bit == '!':
            if key:
                add_args()
                args = []
                key = None

            not_arg = True

        elif bit.startswith('-'):
            if key:
                add_args()
                args = []
                not_arg = False

            key = bit

        else:
            args.append(bit)

    if key:
        add_args()

    if 'extras' in definition:
        definition['extras'] = set(definition['extras'])

    return definition


class IptablesRules(FactBase):
    '''
    Returns a list of iptables rules for a specific table:

    .. code:: python

        [
            {
                'chain': 'PREROUTING',
                'jump': 'DNAT',
                ...
            },
        ]
    '''

    default = list

    def command(self, table='filter'):
        return 'iptables-save -t {0}'.format(table)

    def process(self, output):
        rules = []

        for line in output:
            if line.startswith('-'):
                rules.append(parse_iptables_rule(line))

        return rules


class Ip6tablesRules(IptablesRules):
    '''
    Returns a list of ip6tables rules for a specific table:

    .. code:: python

        [
            {
                'chain': 'PREROUTING',
                'jump': 'DNAT',
                ...
            },
        ]
    '''

    def command(self, table='filter'):
        return 'ip6tables-save -t {0}'.format(table)


class IptablesChains(IptablesRules):
    '''
    Returns a dict of iptables chains & policies:

    .. code:: python

        {
            'NAME': 'POLICY',
        }
    '''

    default = dict

    def process(self, output):
        chains = {}

        for line in output:
            if line.startswith(':'):
                line = line[1:]

                name, policy, _ = line.split()
                chains[name] = policy

        return chains


class Ip6tablesChains(IptablesChains):
    '''
    Returns a dict of ip6tables chains & policies:

    .. code:: python

        {
            'NAME': 'POLICY',
        }
    '''

    def command(self, table='filter'):
        return 'ip6tables-save -t {0}'.format(table)



import re

from pyinfra.api import FactBase


# Valid unit names consist of a "name prefix" and a dot and a suffix specifying the unit type.
# The "unit prefix" must consist of one or more valid characters
# (ASCII letters, digits, ":", "-", "_", ".", and "\").
# The total length of the unit name including the suffix must not exceed 256 characters.
# The type suffix must be one of
# ".service", ".socket", ".device", ".mount", ".automount",
# ".swap", ".target", ".path", ".timer", ".slice", or ".scope".
# Units names can be parameterized by a single argument called the "instance name".
# A template unit must have a single "@" at the end of the name (right before the type suffix).
# The name of the full unit is formed by inserting the instance name
# between "@" and the unit type suffix.
SYSTEMD_UNIT_NAME_REGEX = (
    r'[a-zA-Z0-9\:\-\_\.\\\@]+\.'
    r'(?:service|socket|device|mount|automount|swap|target|path|timer|slice|scope)'
)


class SystemdStatus(FactBase):
    '''
    Returns a dict of name -> status for systemd managed services.
    '''

    command = 'systemctl -al list-units'
    requires_command = 'systemctl'

    default = dict

    regex = r'^({systemd_unit_name_regex})\s+[a-z\-]+\s+[a-z]+\s+([a-z]+)'.format(
        systemd_unit_name_regex=SYSTEMD_UNIT_NAME_REGEX)

    def process(self, output):
        services = {}

        for line in output:
            line = line.strip()
            matches = re.match(self.regex, line)
            if matches:
                is_active = matches.group(2) == 'running' or matches.group(2) == 'waiting'
                services[matches.group(1)] = is_active

        return services


class SystemdEnabled(FactBase):
    '''
    Returns a dict of name -> whether enabled for systemd managed services.
    '''

    command = '''
        systemctl --no-legend -al list-unit-files | while read -r UNIT STATUS; do
            if [ "$STATUS" = generated ] &&
                systemctl is-enabled $UNIT >/dev/null 2>&1; then
                STATUS=enabled
            fi
            echo $UNIT $STATUS
        done
    '''
    requires_command = 'systemctl'

    default = dict

    regex = r'^({systemd_unit_name_regex})\s+([a-z]+)'.format(
        systemd_unit_name_regex=SYSTEMD_UNIT_NAME_REGEX)

    def process(self, output):
        units = {}

        for line in output:
            matches = re.match(self.regex, line)
            if matches:
                units[matches.group(1)] = (
                    matches.group(2) in ('enabled', 'static')
                )

        return units



import re

from collections import defaultdict

from pyinfra.api import FactBase, MaskString, QuoteString, StringCommand
from pyinfra.api.util import try_int

from .util.databases import parse_columns_and_rows


def make_mysql_command(
    database=None,
    user=None,
    password=None,
    host=None,
    port=None,
    executable='mysql',
):
    target_bits = [executable]

    if database:
        target_bits.append(database)

    if user:
        # Quote the username as in may contain special characters
        target_bits.append('-u"{0}"'.format(user))

    if password:
        # Quote the password as it may contain special characters
        target_bits.append(MaskString('-p"{0}"'.format(password)))

    if host:
        target_bits.append('-h{0}'.format(host))

    if port:
        target_bits.append('-P{0}'.format(port))

    return StringCommand(*target_bits)


def make_execute_mysql_command(command, ignore_errors=False, **mysql_kwargs):
    commands_bits = [
        make_mysql_command(**mysql_kwargs),
        '-Be',
        QuoteString(command),  # quote this whole item as a single shell argument
    ]

    if ignore_errors:
        commands_bits.extend(['||', 'true'])

    return StringCommand(*commands_bits)


class MysqlFactBase(FactBase):
    abstract = True

    requires_command = 'mysql'
    ignore_errors = False

    def command(
        self,
        # Details for speaking to MySQL via `mysql` CLI via `mysql` CLI
        mysql_user=None, mysql_password=None,
        mysql_host=None, mysql_port=None,
    ):
        return make_execute_mysql_command(
            self.mysql_command,
            ignore_errors=self.ignore_errors,
            user=mysql_user,
            password=mysql_password,
            host=mysql_host,
            port=mysql_port,
        )


class MysqlDatabases(MysqlFactBase):
    '''
    Returns a dict of existing MySQL databases and associated data:

    .. code:: python

        {
            'mysql': {
                'character_set': 'latin1',
                'collation_name': 'latin1_swedish_ci'
            },
        }
    '''

    default = dict
    mysql_command = 'SELECT * FROM information_schema.SCHEMATA'

    def process(self, output):
        rows = parse_columns_and_rows(
            output, '\t',
            title_parser=lambda title: title.lower(),
        )

        databases = {}

        for details in rows:
            databases[details.pop('schema_name')] = {
                'character_set': details['default_character_set_name'],
                'collation_name': details['default_collation_name'],
            }

        return databases


class MysqlUsers(MysqlFactBase):
    '''
    Returns a dict of MySQL ``user@host``'s and their associated data:

    .. code:: python

        {
            'user@host': {
                'privileges': ['Alter', 'Grant'],
                'max_connections': 5,
                ...
            },
        }
    '''

    default = dict
    mysql_command = 'SELECT * FROM mysql.user'

    @staticmethod
    def process(output):
        rows = parse_columns_and_rows(output, '\t')

        users = {}

        for details in rows:
            if details.get('Host') is None or details.get('User') is None:
                continue  # pragma: no cover

            privileges = []

            for key, value in list(details.items()):
                if key.endswith('_priv') and details.pop(key) == 'Y':
                    privileges.append(key.replace('_priv', ''))

                if key.startswith('max_'):
                    details[key] = try_int(value)

                if key in ('password_expired', 'is_role'):
                    details[key] = value == 'Y'

            details['privileges'] = sorted(privileges)

            # Attach the user in the format user@host
            users['{0}@{1}'.format(
                details.pop('User'), details.pop('Host'),
            )] = details

        return users


MYSQL_GRANT_REGEX = (
    r"^GRANT ([A-Z,\s]+) ON (\*|`[a-z_\\]+`\.\*|'[a-z_]+') "
    r'TO `[A-Z0-9a-z_\-]+`@`[A-Z0-9a-z_\.\-]+`(.*)'
)


class MysqlUserGrants(MysqlFactBase):
    '''
    Returns a dict of ``<database>`.<table>`` with granted privileges for each:

    .. code:: python

        {
            '`pyinfra_stuff`.*': {
                'privileges': [
                    'SELECT',
                    'INSERT'
                ],
                "with_grant_option": false
            },
        }
    '''

    default = dict
    # Ignore errors as SHOW GRANTS will error if the user does not exist
    ignore_errors = True

    def command(
        self, user,
        hostname='localhost',
        # Details for speaking to MySQL via `mysql` CLI via `mysql` CLI
        mysql_user=None, mysql_password=None,
        mysql_host=None, mysql_port=None,
    ):
        self.mysql_command = 'SHOW GRANTS FOR "{0}"@"{1}"'.format(user, hostname)

        return super(MysqlUserGrants, self).command(
            mysql_user, mysql_password,
            mysql_host, mysql_port,
        )

    @staticmethod
    def process(output):
        database_table_privileges = defaultdict(lambda: {
            'privileges': set(),
            'with_grant_option': False,
        })

        for line in output:
            matches = re.match(MYSQL_GRANT_REGEX, line)
            if not matches:
                continue

            privileges, database_table, extras = matches.groups()

            # MySQL outputs this pre-escaped
            database_table = database_table.replace('\\\\', '\\')

            for privilege in privileges.split(','):
                privilege = privilege.strip()
                database_table_privileges[database_table]['privileges'].add(privilege)

            if 'WITH GRANT OPTION' in extras:
                database_table_privileges[database_table]['with_grant_option'] = True

        return database_table_privileges



from os import path
from random import shuffle
from unittest import TestCase

from click.testing import CliRunner

import pyinfra

from pyinfra import pseudo_state
from pyinfra_cli.main import _main, cli

from ..paramiko_util import PatchSSHTestCase


def run_cli(*arguments):
    pyinfra.is_cli = True
    runner = CliRunner(mix_stderr=False)
    result = runner.invoke(cli, arguments)
    pyinfra.is_cli = False
    return result


class TestCliEagerFlags(TestCase):
    def test_print_help(self):
        result = run_cli('--version')
        assert result.exit_code == 0, result.stderr

        result = run_cli('--help')
        assert result.exit_code == 0, result.stderr

    def test_print_facts_list(self):
        result = run_cli('--facts')
        assert result.exit_code == 0, result.stderr

    def test_print_operations_list(self):
        result = run_cli('--operations')
        assert result.exit_code == 0, result.stderr


class TestCliDeployRuns(PatchSSHTestCase):
    def setUp(self):
        pseudo_state.reset()

    def test_invalid_deploy(self):
        result = run_cli(
            '@local',
            'not-a-file.py',
        )
        assert result.exit_code == 1, result.stderr
        assert 'No deploy file: not-a-file.py' in result.stderr

    def test_deploy_inventory(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'server.shell',
            '--debug-data',
        )
        assert result.exit_code == 0, result.stderr

    def test_get_facts(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'fact',
            'os',
        )
        assert result.exit_code == 0, result.stderr
        assert '"somehost": null' in result.stderr

    def test_deploy_operation(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'server.shell',
            'echo hi',
        )
        assert result.exit_code == 0, result.stderr

    def test_deploy_operation_with_all(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventory_all.py'),
            'server.shell',
            'echo hi',
        )
        assert result.exit_code == 0, result.stderr

    def test_exec_command(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'exec',
            '--',
            'echo hi',
        )
        assert result.exit_code == 0, result.stderr

    def test_exec_command_with_options(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'exec',
            '--sudo',
            '--sudo-user', 'pyinfra',
            '--su-user', 'pyinfrawhat',
            '--port', '1022',
            '--user', 'ubuntu',
            '--',
            'echo hi',
        )
        assert result.exit_code == 0, result.stderr

    def test_exec_command_with_serial(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'exec',
            '--serial',
            '--',
            'echo hi',
        )
        assert result.exit_code == 0, result.stderr

    def test_exec_command_with_no_wait(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'exec',
            '--no-wait',
            '--',
            'echo hi',
        )
        assert result.exit_code == 0, result.stderr

    def test_exec_command_with_debug_operations(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'exec',
            '--debug-operations',
            '--',
            'echo hi',
        )
        assert result.exit_code == 0, result.stderr

    def test_exec_command_with_debug_facts(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'exec',
            '--debug-facts',
            '--',
            'echo hi',
        )
        assert result.exit_code == 0, result.stderr

    def test_exec_command_with_debug_data_limit(self):
        result = run_cli(
            path.join('tests', 'deploy', 'inventories', 'inventory.py'),
            'exec',
            '--debug-data',
            '--limit', 'somehost',
            '--',
            'echo hi',
        )
        assert result.exit_code == 0, result.stderr


class TestCliDeployState(PatchSSHTestCase):
    def _assert_op_data(self, correct_op_name_and_host_names):
        state = pseudo_state
        op_order = state.get_op_order()

        assert (
            len(correct_op_name_and_host_names) == len(op_order)
        ), 'Incorrect number of operations detected'

        for i, (correct_op_name, correct_host_names) in enumerate(
            correct_op_name_and_host_names,
        ):
            op_hash = op_order[i]
            op_meta = state.op_meta[op_hash]

            assert list(op_meta['names'])[0] == correct_op_name

            for host in state.inventory:
                op_hashes = state.meta[host]['op_hashes']
                if correct_host_names is True or host.name in correct_host_names:
                    self.assertIn(op_hash, op_hashes)
                else:
                    self.assertNotIn(op_hash, op_hashes)

    def test_deploy(self):
        task_file_path = path.join('tests', 'deploy', 'tasks', 'a_task.py')
        nested_task_path = path.join('tests', 'deploy', 'tasks', 'another_task.py')
        correct_op_name_and_host_names = [
            ('First main operation', True),  # true for all hosts
            ('Second main operation', ('somehost',)),
            ('{0} | First task operation'.format(task_file_path), ('anotherhost',)),
            ('{0} | Task order loop 1'.format(task_file_path), ('anotherhost',)),
            ('{0} | 2nd Task order loop 1'.format(task_file_path), ('anotherhost',)),
            ('{0} | Task order loop 2'.format(task_file_path), ('anotherhost',)),
            ('{0} | 2nd Task order loop 2'.format(task_file_path), ('anotherhost',)),
            (
                '{0} | {1} | Second task operation'.format(task_file_path, nested_task_path),
                ('anotherhost',),
            ),
            ('{0} | First task operation'.format(task_file_path), True),
            ('{0} | Task order loop 1'.format(task_file_path), True),
            ('{0} | 2nd Task order loop 1'.format(task_file_path), True),
            ('{0} | Task order loop 2'.format(task_file_path), True),
            ('{0} | 2nd Task order loop 2'.format(task_file_path), True),
            ('{0} | {1} | Second task operation'.format(task_file_path, nested_task_path), True),
            ('My deploy | First deploy operation', True),
            ('My deploy | My nested deploy | First nested deploy operation', True),
            ('My deploy | Second deploy operation', True),
            ('Loop-0 main operation', True),
            ('Loop-1 main operation', True),
            ('Third main operation', True),
            ('Order loop 1', True),
            ('2nd Order loop 1', True),
            ('Order loop 2', True),
            ('2nd Order loop 2', True),
            ('Final limited operation', ('somehost',)),
        ]

        # Run 3 iterations of the test - each time shuffling the order of the
        # hosts - ensuring that the ordering has no effect on the operation order.
        for _ in range(3):
            pseudo_state.reset()

            hosts = ['somehost', 'anotherhost', 'someotherhost']
            shuffle(hosts)

            result = run_cli(
                ','.join(hosts),
                path.join('tests', 'deploy', 'deploy.py'),
            )
            assert result.exit_code == 0, result.stderr

            self._assert_op_data(correct_op_name_and_host_names)

    def test_legacy_deploy(self):
        pseudo_state.reset()

        result = run_cli(
            'somehost',
            path.join('tests', 'deploy', 'deploy_legacy.py'),
        )
        assert result.exit_code == 0, result.stderr

        correct_op_name_and_host_names = [
            ('First main operation', True),
            ('My deploy | My nested deploy | First nested deploy operation', True),
            ('My deploy | Second deploy operation', True),
        ]

        self._assert_op_data(correct_op_name_and_host_names)

    def test_interdependent_deploy(self):
        pseudo_state.reset()

        result = run_cli(
            'somehost',
            path.join('tests', 'deploy', 'deploy_interdependent.py'),
        )
        assert result.exit_code == 0, result.stderr

        # Check every operation had commands/changes - this ensures that each
        # combo (add/remove/add) always had changes.
        for host, ops in pseudo_state.ops.items():
            for _, op in ops.items():
                assert len(op['commands']) > 0


class TestDirectMainExecution(PatchSSHTestCase):
    '''
    These tests are very similar as above, without the click wrappers - basically
    here because coverage.py fails to properly detect all the code under the wrapper.
    '''

    def test_deploy_operation_direct(self):
        with self.assertRaises(SystemExit) as e:
            _main(
                inventory=path.join('tests', 'test_deploy', 'inventories', 'inventory.py'),
                operations=['server.shell', 'echo hi'],
                verbosity=0, user=None, port=None, key=None, key_password=None,
                password=None, sudo=False, sudo_user=None, use_sudo_password=False, su_user=None,
                parallel=None, fail_percent=0, dry=False, limit=None, no_wait=False, serial=False,
                winrm_username=None, winrm_password=None, winrm_port=None,
                shell_executable=None, quiet=False,
                debug=False, debug_data=False, debug_facts=False, debug_operations=False,
            )
            assert e.args == (0,)



# encoding: utf-8

from __future__ import unicode_literals

from socket import (
    error as socket_error,
    gaierror,
)
from unittest import TestCase

from mock import call, MagicMock, mock_open, patch
from paramiko import (
    AuthenticationException,
    PasswordRequiredException,
    SSHException,
)

import pyinfra

from pyinfra.api import Config, MaskString, State, StringCommand
from pyinfra.api.connect import connect_all
from pyinfra.api.connectors.ssh import _get_sftp_connection
from pyinfra.api.exceptions import ConnectError, PyinfraError

from ..util import make_inventory


@patch('pyinfra.api.connectors.ssh.SSHClient.get_transport', MagicMock())
@patch('pyinfra.api.connectors.ssh.open', mock_open(read_data='test!'), create=True)
class TestSSHConnector(TestCase):
    def setUp(self):
        self.fake_connect_patch = patch('pyinfra.api.connectors.ssh.SSHClient.connect')
        self.fake_connect_mock = self.fake_connect_patch.start()
        _get_sftp_connection.cache = {}

    def tearDown(self):
        self.fake_connect_patch.stop()

    def test_connect_all(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        connect_all(state)
        assert len(state.active_hosts) == 2

    def test_connect_host(self):
        inventory = make_inventory()
        state = State(inventory, Config())
        host = inventory.get_host('somehost')
        host.connect(reason=True)
        assert len(state.active_hosts) == 0

    def test_connect_all_password(self):
        inventory = make_inventory(ssh_password='test')

        # Get a host
        somehost = inventory.get_host('somehost')
        assert somehost.data.ssh_password == 'test'

        state = State(inventory, Config())
        connect_all(state)

        assert len(state.active_hosts) == 2

    @patch('pyinfra.api.connectors.ssh.path.isfile', lambda *args, **kwargs: True)
    @patch('pyinfra.api.connectors.ssh.RSAKey.from_private_key_file')
    def test_connect_exceptions(self, fake_key_open):
        for exception_class in (
            AuthenticationException,
            SSHException,
            gaierror,
            socket_error,
            EOFError,
        ):
            state = State(make_inventory(hosts=(
                ('somehost', {'ssh_key': 'testkey'}),
            )), Config())

            def raise_exception(*args, **kwargs):
                raise exception_class

            self.fake_connect_mock.side_effect = raise_exception

            with self.assertRaises(PyinfraError):
                connect_all(state)

            assert len(state.active_hosts) == 0

    def test_connect_with_rsa_ssh_key(self):
        state = State(make_inventory(hosts=(
            ('somehost', {'ssh_key': 'testkey'}),
        )), Config())

        with patch('pyinfra.api.connectors.ssh.path.isfile', lambda *args, **kwargs: True), \
                patch('pyinfra.api.connectors.ssh.RSAKey.from_private_key_file') as fake_key_open:

            fake_key = MagicMock()
            fake_key_open.return_value = fake_key

            state.deploy_dir = '/'

            connect_all(state)

            # Check the key was created properly
            fake_key_open.assert_called_with(filename='testkey')
            # Check the certificate file was then loaded
            fake_key.load_certificate.assert_called_with('testkey.pub')

            # And check the Paramiko SSH call was correct
            self.fake_connect_mock.assert_called_with(
                'somehost',
                allow_agent=False,
                look_for_keys=False,
                pkey=fake_key,
                timeout=10,
                username='vagrant',
            )

        # Check that loading the same key again is cached in the state
        second_state = State(make_inventory(hosts=(
            ('somehost', {'ssh_key': 'testkey'}),
        )), Config())
        second_state.private_keys = state.private_keys

        connect_all(second_state)

    def test_connect_with_rsa_ssh_key_password(self):
        state = State(make_inventory(hosts=(
            ('somehost', {'ssh_key': 'testkey', 'ssh_key_password': 'testpass'}),
        )), Config())

        with patch(
            'pyinfra.api.connectors.ssh.path.isfile',
            lambda *args, **kwargs: True,
        ), patch(
            'pyinfra.api.connectors.ssh.RSAKey.from_private_key_file',
        ) as fake_key_open:
            fake_key = MagicMock()

            def fake_key_open_fail(*args, **kwargs):
                if 'password' not in kwargs:
                    raise PasswordRequiredException()
                return fake_key

            fake_key_open.side_effect = fake_key_open_fail

            state.deploy_dir = '/'

            connect_all(state)

            # Check the key was created properly
            fake_key_open.assert_called_with(filename='testkey', password='testpass')
            # Check the certificate file was then loaded
            fake_key.load_certificate.assert_called_with('testkey.pub')

    def test_connect_with_rsa_ssh_key_password_from_prompt(self):
        state = State(make_inventory(hosts=(
            ('somehost', {'ssh_key': 'testkey'}),
        )), Config())

        with patch(
            'pyinfra.api.connectors.ssh.path.isfile',
            lambda *args, **kwargs: True,
        ), patch(
            'pyinfra.api.connectors.ssh.getpass',
            lambda *args, **kwargs: 'testpass',
        ), patch(
            'pyinfra.api.connectors.ssh.RSAKey.from_private_key_file',
        ) as fake_key_open:
            fake_key = MagicMock()

            def fake_key_open_fail(*args, **kwargs):
                if 'password' not in kwargs:
                    raise PasswordRequiredException()
                return fake_key

            fake_key_open.side_effect = fake_key_open_fail

            state.deploy_dir = '/'

            pyinfra.is_cli = True
            connect_all(state)
            pyinfra.is_cli = False

            # Check the key was created properly
            fake_key_open.assert_called_with(filename='testkey', password='testpass')
            # Check the certificate file was then loaded
            fake_key.load_certificate.assert_called_with('testkey.pub')

    def test_connect_with_rsa_ssh_key_missing_password(self):
        state = State(make_inventory(hosts=(
            ('somehost', {'ssh_key': 'testkey'}),
        )), Config())

        with patch(
            'pyinfra.api.connectors.ssh.path.isfile',
            lambda *args, **kwargs: True,
        ), patch(
            'pyinfra.api.connectors.ssh.RSAKey.from_private_key_file',
        ) as fake_key_open:

            def fake_key_open_fail(*args, **kwargs):
                raise PasswordRequiredException

            fake_key_open.side_effect = fake_key_open_fail

            fake_key = MagicMock()
            fake_key_open.return_value = fake_key

            state.deploy_dir = '/'

            with self.assertRaises(PyinfraError) as e:
                connect_all(state)

            assert e.exception.args[0] == (
                'Private key file (testkey) is encrypted, set ssh_key_password '
                'to use this key'
            )

    def test_connect_with_rsa_ssh_key_wrong_password(self):
        state = State(make_inventory(hosts=(
            ('somehost', {'ssh_key': 'testkey', 'ssh_key_password': 'testpass'}),
        )), Config())

        with patch(
            'pyinfra.api.connectors.ssh.path.isfile',
            lambda *args, **kwargs: True,
        ), patch(
            'pyinfra.api.connectors.ssh.RSAKey.from_private_key_file',
        ) as fake_key_open:

            def fake_key_open_fail(*args, **kwargs):
                if 'password' not in kwargs:
                    raise PasswordRequiredException
                raise SSHException

            fake_key_open.side_effect = fake_key_open_fail

            fake_key = MagicMock()
            fake_key_open.return_value = fake_key

            state.deploy_dir = '/'

            with self.assertRaises(PyinfraError) as e:
                connect_all(state)

            assert e.exception.args[0] == 'Incorrect password for private key: testkey'

    def test_connect_with_dss_ssh_key(self):
        state = State(make_inventory(hosts=(
            ('somehost', {'ssh_key': 'testkey'}),
        )), Config())

        with patch('pyinfra.api.connectors.ssh.path.isfile', lambda *args, **kwargs: True), \
                patch('pyinfra.api.connectors.ssh.RSAKey.from_private_key_file') as fake_rsa_key_open, \
                patch('pyinfra.api.connectors.ssh.DSSKey.from_private_key_file') as fake_key_open:  # noqa

            def fake_rsa_key_open_fail(*args, **kwargs):
                raise SSHException

            fake_rsa_key_open.side_effect = fake_rsa_key_open_fail

            fake_key = MagicMock()
            fake_key_open.return_value = fake_key

            state.deploy_dir = '/'

            connect_all(state)

            # Check the key was created properly
            fake_key_open.assert_called_with(filename='testkey')

            # And check the Paramiko SSH call was correct
            self.fake_connect_mock.assert_called_with(
                'somehost',
                allow_agent=False,
                look_for_keys=False,
                pkey=fake_key,
                timeout=10,
                username='vagrant',
            )

        # Check that loading the same key again is cached in the state
        second_state = State(make_inventory(hosts=(
            ('somehost', {'ssh_key': 'testkey'}),
        )), Config())
        second_state.private_keys = state.private_keys

        connect_all(second_state)

    def test_connect_with_missing_ssh_key(self):
        state = State(make_inventory(hosts=(
            ('somehost', {'ssh_key': 'testkey'}),
        )), Config())

        with self.assertRaises(IOError) as e:
            connect_all(state)

        # Ensure pyinfra style IOError
        self.assertTrue(e.exception.args[0].startswith('No such private key file:'))

    @patch('pyinfra.api.connectors.ssh.SSHClient')
    def test_run_shell_command(self, fake_ssh_client):
        fake_ssh = MagicMock()
        fake_stdin = MagicMock()
        fake_stdout = MagicMock()
        fake_ssh.exec_command.return_value = fake_stdin, fake_stdout, MagicMock()

        fake_ssh_client.return_value = fake_ssh

        inventory = make_inventory(hosts=('somehost',))
        State(inventory, Config())
        host = inventory.get_host('somehost')
        host.connect()

        command = 'echo Šablony'
        fake_stdout.channel.recv_exit_status.return_value = 0

        out = host.run_shell_command(command, stdin='hello', print_output=True)
        assert len(out) == 3

        status, stdout, stderr = out
        assert status is True
        fake_stdin.write.assert_called_with(b'hello\n')

        combined_out = host.run_shell_command(
            command, stdin='hello', print_output=True,
            return_combined_output=True,
        )
        assert len(combined_out) == 2

        fake_ssh.exec_command.assert_called_with("sh -c 'echo Šablony'", get_pty=False)

    @patch('pyinfra.api.connectors.ssh.click')
    @patch('pyinfra.api.connectors.ssh.SSHClient')
    def test_run_shell_command_masked(self, fake_ssh_client, fake_click):
        fake_ssh = MagicMock()
        fake_stdout = MagicMock()
        fake_ssh.exec_command.return_value = MagicMock(), fake_stdout, MagicMock()

        fake_ssh_client.return_value = fake_ssh

        inventory = make_inventory(hosts=('somehost',))
        State(inventory, Config())
        host = inventory.get_host('somehost')
        host.connect()

        command = StringCommand('echo', MaskString('top-secret-stuff'))
        fake_stdout.channel.recv_exit_status.return_value = 0

        out = host.run_shell_command(command, print_output=True, print_input=True)
        assert len(out) == 3

        status, stdout, stderr = out
        assert status is True

        fake_ssh.exec_command.assert_called_with(
            "sh -c 'echo top-secret-stuff'",
            get_pty=False,
        )

        fake_click.echo.assert_called_with(
            "{0}>>> sh -c 'echo ***'".format(host.print_prefix),
            err=True,
        )

    @patch('pyinfra.api.connectors.ssh.SSHClient')
    def test_run_shell_command_success_exit_code(self, fake_ssh_client):
        fake_ssh = MagicMock()
        fake_stdin = MagicMock()
        fake_stdout = MagicMock()
        fake_ssh.exec_command.return_value = fake_stdin, fake_stdout, MagicMock()

        fake_ssh_client.return_value = fake_ssh

        inventory = make_inventory(hosts=('somehost',))
        State(inventory, Config())
        host = inventory.get_host('somehost')
        host.connect()

        command = 'echo hi'
        fake_stdout.channel.recv_exit_status.return_value = 1

        out = host.run_shell_command(command, success_exit_codes=[1])
        assert len(out) == 3
        assert out[0] is True

    @patch('pyinfra.api.connectors.ssh.SSHClient')
    def test_run_shell_command_error(self, fake_ssh_client):
        fake_ssh = MagicMock()
        fake_stdin = MagicMock()
        fake_stdout = MagicMock()
        fake_ssh.exec_command.return_value = fake_stdin, fake_stdout, MagicMock()

        fake_ssh_client.return_value = fake_ssh

        inventory = make_inventory(hosts=('somehost',))
        state = State(inventory, Config())
        host = inventory.get_host('somehost')
        host.connect(state)

        command = 'echo hi'
        fake_stdout.channel.recv_exit_status.return_value = 1

        out = host.run_shell_command(command)
        assert len(out) == 3
        assert out[0] is False

    @patch('pyinfra.api.connectors.ssh.SSHClient')
    @patch('pyinfra.api.connectors.ssh.SFTPClient')
    def test_put_file(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=('anotherhost',))
        State(inventory, Config())
        host = inventory.get_host('anotherhost')
        host.connect()

        fake_open = mock_open(read_data='test!')
        with patch('pyinfra.api.util.open', fake_open, create=True):
            status = host.put_file(
                'not-a-file', 'not-another-file',
                print_output=True,
            )

        assert status is True
        fake_sftp_client.from_transport().putfo.assert_called_with(
            fake_open(), 'not-another-file',
        )

    @patch('pyinfra.api.connectors.ssh.SSHClient')
    @patch('pyinfra.api.connectors.ssh.SFTPClient')
    def test_put_file_sudo(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=('anotherhost',))
        State(inventory, Config())
        host = inventory.get_host('anotherhost')
        host.connect()

        stdout_mock = MagicMock()
        stdout_mock.channel.recv_exit_status.return_value = 0
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        fake_open = mock_open(read_data='test!')
        with patch('pyinfra.api.util.open', fake_open, create=True):
            status = host.put_file(
                'not-a-file', 'not-another-file',
                print_output=True,
                sudo=True,
                sudo_user='ubuntu',
            )

        assert status is True

        fake_ssh_client().exec_command.assert_called_with((
            "sudo -H -n -u ubuntu sh -c 'mv "
            '/tmp/pyinfra-43db9984686317089fefcf2e38de527e4cb44487 '
            "not-another-file && chown ubuntu not-another-file'"
        ), get_pty=False)

        fake_sftp_client.from_transport().putfo.assert_called_with(
            fake_open(), '/tmp/pyinfra-43db9984686317089fefcf2e38de527e4cb44487',
        )

    @patch('pyinfra.api.connectors.ssh.SSHClient')
    @patch('pyinfra.api.connectors.ssh.SFTPClient')
    def test_put_file_su_user_fail(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=('anotherhost',))
        State(inventory, Config())
        host = inventory.get_host('anotherhost')
        host.connect()

        stdout_mock = MagicMock()
        stdout_mock.channel.recv_exit_status.return_value = 1
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        fake_open = mock_open(read_data='test!')
        with patch('pyinfra.api.util.open', fake_open, create=True):
            status = host.put_file(
                'not-a-file', 'not-another-file',
                print_output=True,
                su_user='centos',
            )

        assert status is False

        fake_ssh_client().exec_command.assert_called_with((
            "su centos -c 'sh -c '\"'\"'mv "
            '/tmp/pyinfra-43db9984686317089fefcf2e38de527e4cb44487 '
            "not-another-file && chown centos not-another-file'\"'\"''"
        ), get_pty=False)

        fake_sftp_client.from_transport().putfo.assert_called_with(
            fake_open(), '/tmp/pyinfra-43db9984686317089fefcf2e38de527e4cb44487',
        )

    @patch('pyinfra.api.connectors.ssh.SSHClient')
    @patch('pyinfra.api.connectors.ssh.SFTPClient')
    def test_get_file(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=('somehost',))
        State(inventory, Config())
        host = inventory.get_host('somehost')
        host.connect()

        fake_open = mock_open(read_data='test!')
        with patch('pyinfra.api.util.open', fake_open, create=True):
            status = host.get_file(
                'not-a-file', 'not-another-file',
                print_output=True,
            )

        assert status is True
        fake_sftp_client.from_transport().getfo.assert_called_with(
            'not-a-file', fake_open(),
        )

    @patch('pyinfra.api.connectors.ssh.SSHClient')
    @patch('pyinfra.api.connectors.ssh.SFTPClient')
    def test_get_file_sudo(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=('somehost',))
        State(inventory, Config())
        host = inventory.get_host('somehost')
        host.connect()

        stdout_mock = MagicMock()
        stdout_mock.channel.recv_exit_status.return_value = 0
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        fake_open = mock_open(read_data='test!')
        with patch('pyinfra.api.util.open', fake_open, create=True):
            status = host.get_file(
                'not-a-file', 'not-another-file',
                print_output=True,
                sudo=True,
                sudo_user='ubuntu',
            )

        assert status is True

        fake_ssh_client().exec_command.assert_has_calls([
            call((
                "sudo -H -n -u ubuntu sh -c 'cp not-a-file "
                "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508 && chmod +r not-a-file'"
            ), get_pty=False),
            call((
                "sudo -H -n -u ubuntu sh -c 'rm -f "
                "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508'"
            ), get_pty=False),
        ])

        fake_sftp_client.from_transport().getfo.assert_called_with(
            '/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508', fake_open(),
        )

    @patch('pyinfra.api.connectors.ssh.SSHClient')
    def test_get_file_sudo_copy_fail(self, fake_ssh_client):
        inventory = make_inventory(hosts=('somehost',))
        State(inventory, Config())
        host = inventory.get_host('somehost')
        host.connect()

        stdout_mock = MagicMock()
        stdout_mock.channel.recv_exit_status.return_value = 1
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        status = host.get_file(
            'not-a-file', 'not-another-file',
            print_output=True,
            sudo=True,
            sudo_user='ubuntu',
        )

        assert status is False

        fake_ssh_client().exec_command.assert_has_calls([
            call((
                "sudo -H -n -u ubuntu sh -c 'cp not-a-file "
                "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508 && chmod +r not-a-file'"
            ), get_pty=False),
        ])

    @patch('pyinfra.api.connectors.ssh.SSHClient')
    @patch('pyinfra.api.connectors.ssh.SFTPClient')
    def test_get_file_sudo_remove_fail(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=('somehost',))
        State(inventory, Config())
        host = inventory.get_host('somehost')
        host.connect()

        stdout_mock = MagicMock()
        stdout_mock.channel.recv_exit_status.side_effect = [0, 1]
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        fake_open = mock_open(read_data='test!')
        with patch('pyinfra.api.util.open', fake_open, create=True):
            status = host.get_file(
                'not-a-file', 'not-another-file',
                print_output=True,
                sudo=True,
                sudo_user='ubuntu',
            )

        assert status is False

        fake_ssh_client().exec_command.assert_has_calls([
            call((
                "sudo -H -n -u ubuntu sh -c 'cp not-a-file "
                "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508 && chmod +r not-a-file'"
            ), get_pty=False),
            call((
                "sudo -H -n -u ubuntu sh -c 'rm -f "
                "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508'"
            ), get_pty=False),
        ])

        fake_sftp_client.from_transport().getfo.assert_called_with(
            '/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508', fake_open(),
        )

    @patch('pyinfra.api.connectors.ssh.SSHClient')
    @patch('pyinfra.api.connectors.ssh.SFTPClient')
    def test_get_file_su_user(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=('somehost',))
        State(inventory, Config())
        host = inventory.get_host('somehost')
        host.connect()

        stdout_mock = MagicMock()
        stdout_mock.channel.recv_exit_status.return_value = 0
        fake_ssh_client().exec_command.return_value = MagicMock(), stdout_mock, MagicMock()

        fake_open = mock_open(read_data='test!')
        with patch('pyinfra.api.util.open', fake_open, create=True):
            status = host.get_file(
                'not-a-file', 'not-another-file',
                print_output=True,
                su_user='centos',
            )

        assert status is True

        fake_ssh_client().exec_command.assert_has_calls([
            call((
                "su centos -c 'sh -c '\"'\"'cp not-a-file "
                '/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508 && chmod +r '
                "not-a-file'\"'\"''"
            ), get_pty=False),
            call((
                "su centos -c 'sh -c '\"'\"'rm -f "
                "/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508'\"'\"''"
            ), get_pty=False),
        ])

        fake_sftp_client.from_transport().getfo.assert_called_with(
            '/tmp/pyinfra-e9c0d3c8ffca943daa0e75511b0a09c84b59c508', fake_open(),
        )

    @patch('pyinfra.api.connectors.ssh.SSHClient')
    @patch('pyinfra.api.connectors.ssh.SFTPClient')
    def test_get_sftp_fail(self, fake_sftp_client, fake_ssh_client):
        inventory = make_inventory(hosts=('anotherhost',))
        State(inventory, Config())
        host = inventory.get_host('anotherhost')
        host.connect()

        def raise_exception(*args, **kwargs):
            raise SSHException()

        fake_sftp_client.from_transport.side_effect = raise_exception

        fake_open = mock_open(read_data='test!')
        with patch('pyinfra.api.util.open', fake_open, create=True):
            with self.assertRaises(ConnectError):
                host.put_file(
                    'not-a-file', 'not-another-file',
                    print_output=True,
                )



from pyinfra.api import FactBase

from .util import make_cat_files_command
from .util.packaging import parse_zypper_repositories


class ZypperRepositories(FactBase):
    '''
    Returns a list of installed zypper repositories:

    .. code:: python

        [
            {
                'name': 'Main Repository',
                'enabled': '1',
                'autorefresh': '0',
                'baseurl': 'http://download.opensuse.org/distribution/leap/$releasever/repo/oss/',
                'type': 'rpm-md',
            },
        ]
    '''

    command = make_cat_files_command(
        '/etc/zypp/repos.d/*.repo',
    )
    requires_command = 'zypper'

    default = list

    def process(self, output):
        return parse_zypper_repositories(output)



from __future__ import division, print_function, unicode_literals

import re

from functools import wraps
from hashlib import sha1
from inspect import getframeinfo, stack
from os import path
from socket import (
    error as socket_error,
    timeout as timeout_error,
)
from types import GeneratorType

import click
import six

from jinja2 import (
    Environment,
    StrictUndefined,
    TemplateSyntaxError,
    UndefinedError,
)
from paramiko import SSHException

from pyinfra import logger

from .exceptions import PyinfraError

# 64kb chunks
BLOCKSIZE = 65536

# Caches
TEMPLATES = {}
FILE_SHAS = {}

PYINFRA_API_DIR = path.dirname(__file__)


def get_kwargs_str(kwargs):
    if not kwargs:
        return ''

    items = [
        '{0}={1}'.format(key, value)
        for key, value in kwargs.items()
        if key not in ('self', 'state', 'host')
    ]
    return '({0})'.format(', '.join(items))


def try_int(value):
    try:
        return int(value)
    except (TypeError, ValueError):
        return value


def memoize(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        key = '{0}{1}'.format(args, kwargs)
        if key in wrapper.cache:
            return wrapper.cache[key]

        value = func(*args, **kwargs)
        wrapper.cache[key] = value
        return value

    wrapper.cache = {}
    return wrapper


def get_call_location():
    frame = get_caller_frameinfo(frame_offset=1)  # escape *this* function
    return 'line {0} in {1}'.format(
        frame.lineno,
        path.relpath(frame.filename),
    )


def get_caller_frameinfo(frame_offset=0):
    # Default frames to look at is 2; one for this function call itself
    # in util.py and one for the caller of this function within pyinfra
    # giving the external call frame (ie end user deploy code).
    frame_shift = 2 + frame_offset

    stack_items = stack()

    frame = stack_items[frame_shift][0]
    info = getframeinfo(frame)

    del stack_items

    return info


def get_operation_order_from_stack(state):
    stack_items = list(reversed(stack()))

    # Find the *first* occurrence of our deploy file in the reversed stack
    if state.current_deploy_filename:
        for i, stack_item in enumerate(stack_items):
            frame = getframeinfo(stack_item[0])
            if frame.filename == state.current_deploy_filename:
                break
    else:
        i = 0

    # Now generate a list of line numbers *following that file*
    line_numbers = []
    for stack_item in stack_items[i:]:
        frame = getframeinfo(stack_item[0])

        if frame.filename.startswith(PYINFRA_API_DIR):
            continue

        if state.loop_filename and frame.filename == state.loop_filename:
            line_numbers.extend([state.loop_line, state.loop_counter])

        line_numbers.append(frame.lineno)

    del stack_items

    return line_numbers


def extract_callable_datas(datas):
    for data in datas:
        # Support for dynamic data, ie @deploy wrapped data defaults where
        # the data is stored on the state temporarily.
        if callable(data):
            data = data()

        yield data


class FallbackDict(object):
    '''
    Combines multiple AttrData's to search for attributes.
    '''

    override_datas = None

    def __init__(self, *datas):
        datas = list(datas)

        # Inject an empty override data so we can assign during deploy
        self.__dict__['override_datas'] = {}
        datas.insert(0, self.override_datas)

        self.__dict__['datas'] = tuple(datas)

    def __getattr__(self, key):
        for data in extract_callable_datas(self.datas):
            if key in data:
                return data[key]

    def __setattr__(self, key, value):
        self.override_datas[key] = value

    def __str__(self):
        return six.text_type(self.datas)

    def dict(self):
        out = {}

        # Copy and reverse data objects (such that the first items override
        # the last, matching __getattr__ output).
        datas = list(self.datas)
        datas.reverse()

        for data in extract_callable_datas(datas):
            out.update(data)

        return out


def unroll_generators(generator):
    '''
    Take a generator and unroll any sub-generators recursively. This is
    essentially a Python 2 way of doing `yield from` in Python 3 (given
    iterating the entire thing).
    '''

    # Ensure we have a generator (prevents ccommands returning lists)
    if not isinstance(generator, GeneratorType):
        raise TypeError('{0} is not a generator'.format(generator))

    items = []

    for item in generator:
        if isinstance(item, GeneratorType):
            items.extend(unroll_generators(item))
        else:
            items.append(item)

    return items


def get_template(filename_or_string, is_string=False):
    '''
    Gets a jinja2 ``Template`` object for the input filename or string, with caching
    based on the filename of the template, or the SHA1 of the input string.
    '''

    # Cache against string sha or just the filename
    cache_key = sha1_hash(filename_or_string) if is_string else filename_or_string

    if cache_key in TEMPLATES:
        return TEMPLATES[cache_key]

    if is_string:
        # Set the input string as our template
        template_string = filename_or_string

    else:
        # Load template data into memory
        with open(filename_or_string, 'r') as file_io:
            template_string = file_io.read()

    TEMPLATES[cache_key] = Environment(
        undefined=StrictUndefined,
        keep_trailing_newline=True,
    ).from_string(template_string)

    return TEMPLATES[cache_key]


def underscore(name):
    '''
    Transform CamelCase -> snake_case.
    '''

    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def sha1_hash(string):
    '''
    Return the SHA1 of the input string.
    '''

    hasher = sha1()
    hasher.update(string.encode('utf-8'))
    return hasher.hexdigest()


def format_exception(e):
    return '{0}{1}'.format(e.__class__.__name__, e.args)


def print_host_combined_output(host, combined_output_lines):
    for type_, line in combined_output_lines:
        if type_ == 'stderr':
            logger.error('{0}{1}'.format(
                host.print_prefix,
                click.style(line, 'red'),
            ))
        else:
            logger.error('{0}{1}'.format(
                host.print_prefix,
                line,
            ))


def log_error_or_warning(host, ignore_errors, description=''):
    log_func = logger.error
    log_color = 'red'
    log_text = 'Error: ' if description else 'Error'

    if ignore_errors:
        log_func = logger.warning
        log_color = 'yellow'
        log_text = 'Error (ignored): ' if description else 'Error (ignored)'

    log_func('{0}{1}{2}'.format(
        host.print_prefix,
        click.style(log_text, log_color),
        description,
    ))


def log_host_command_error(host, e, timeout=0):
    if isinstance(e, timeout_error):
        logger.error('{0}{1}'.format(
            host.print_prefix,
            click.style('Command timed out after {0}s'.format(
                timeout,
            ), 'red'),
        ))

    elif isinstance(e, (socket_error, SSHException)):
        logger.error('{0}{1}'.format(
            host.print_prefix,
            click.style('Command socket/SSH error: {0}'.format(
                format_exception(e)), 'red',
            ),
        ))

    elif isinstance(e, IOError):
        logger.error('{0}{1}'.format(
            host.print_prefix,
            click.style('Command IO error: {0}'.format(
                format_exception(e)), 'red',
            ),
        ))

    # Still here? Re-raise!
    else:
        raise e


@memoize
def show_get_arg_value_warning():
    logger.warning((
        'Use of jinja2 templating in arguments is deprecated, '
        'please use builtin Python string formatting.'
    ))


# TODO: remove this! COMPAT w/<1
def get_arg_value(state, host, arg):
    '''
    This functions is **deprecated**.
    '''

    if isinstance(arg, six.string_types):
        data = {
            'host': host,
            'inventory': state.inventory,
        }

        try:
            rendered_arg = get_template(arg, is_string=True).render(data)
        except (TemplateSyntaxError, UndefinedError) as e:
            raise PyinfraError('Error in template string: {0}'.format(e))
        else:
            if rendered_arg != arg:
                show_get_arg_value_warning()
            return rendered_arg

    elif isinstance(arg, list):
        return [get_arg_value(state, host, value) for value in arg]

    elif isinstance(arg, tuple):
        return tuple(get_arg_value(state, host, value) for value in arg)

    elif isinstance(arg, dict):
        return {
            key: get_arg_value(state, host, value)
            for key, value in six.iteritems(arg)
        }

    return arg


def make_hash(obj):
    '''
    Make a hash from an arbitrary nested dictionary, list, tuple or set, used to generate
    ID's for operations based on their name & arguments.
    '''

    if isinstance(obj, (set, tuple, list)):
        hash_string = ''.join([make_hash(e) for e in obj])

    elif isinstance(obj, dict):
        hash_string = ''.join(
            ''.join((key, make_hash(value)))
            for key, value in six.iteritems(obj)
        )

    else:
        hash_string = (
            # Capture integers first (as 1 == True)
            '{0}'.format(obj) if isinstance(obj, int)
            # Constants - the values can change between hosts but we should still
            # group them under the same operation hash.
            else '_PYINFRA_CONSTANT' if obj in (True, False, None)
            # Plain strings
            else obj if isinstance(obj, six.string_types)
            # Objects with __name__s
            else obj.__name__ if hasattr(obj, '__name__')
            # Objects with names
            else obj.name if hasattr(obj, 'name')
            # Repr anything else
            else repr(obj)
        )

    return sha1_hash(hash_string)


class get_file_io(object):
    '''
    Given either a filename or an existing IO object, this context processor
    will open and close filenames, and leave IO objects alone.
    '''

    close = False

    def __init__(self, filename_or_io, mode='rb'):
        if not (
            # Check we can be read
            hasattr(filename_or_io, 'read')
            # Or we're a filename
            or isinstance(filename_or_io, six.string_types)
        ):
            raise TypeError('Invalid filename or IO object: {0}'.format(
                filename_or_io,
            ))

        self.filename_or_io = filename_or_io
        self.mode = mode

    def __enter__(self):
        # If we have a read attribute, just use the object as-is
        if hasattr(self.filename_or_io, 'read'):
            file_io = self.filename_or_io

        # Otherwise, assume a filename and open it up
        else:
            file_io = open(self.filename_or_io, self.mode)

            # Attach to self for closing on __exit__
            self.file_io = file_io
            self.close = True

        # Ensure we're at the start of the file
        file_io.seek(0)
        return file_io

    def __exit__(self, type, value, traceback):
        if self.close:
            self.file_io.close()

    @property
    def cache_key(self):
        # If we're a filename, cache against that - we don't cache in-memory
        # file objects.
        if isinstance(self.filename_or_io, six.string_types):
            return self.filename_or_io


def get_file_sha1(filename_or_io):
    '''
    Calculates the SHA1 of a file or file object using a buffer to handle larger files.
    '''

    file_data = get_file_io(filename_or_io)
    cache_key = file_data.cache_key

    if cache_key and cache_key in FILE_SHAS:
        return FILE_SHAS[cache_key]

    with file_data as file_io:
        hasher = sha1()
        buff = file_io.read(BLOCKSIZE)

        while len(buff) > 0:
            if isinstance(buff, six.text_type):
                buff = buff.encode('utf-8')

            hasher.update(buff)
            buff = file_io.read(BLOCKSIZE)

    digest = hasher.hexdigest()

    if cache_key:
        FILE_SHAS[cache_key] = digest

    return digest



'''
The windows_files module handles windows filesystem state, file uploads and template generation.
'''

from __future__ import unicode_literals

import ntpath

import six

from pyinfra import logger
from pyinfra.api import operation, OperationError, OperationTypeError


@operation(pipeline_facts={
    'windows_file': 'name',
})
def file(
    name,
    present=True, assume_present=False,
    user=None, group=None, mode=None, touch=False,
    create_remote_dir=True,
    state=None, host=None,
):
    '''
    Add/remove/update files.

    + name: name/path of the remote file
    + present: whether the file should exist
    + assume_present: whether to assume the file exists
    + TODO: user: user to own the files
    + TODO: group: group to own the files
    + TODO: mode: permissions of the files as an integer, eg: 755
    + touch: whether to touch the file
    + create_remote_dir: create the remote directory if it doesn't exist

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Example:

    .. code:: python

        files.file(
            name='Create c:\\temp\\hello.txt',
            path='c:\\temp\\hello.txt',
            touch=True,
        )
    '''

    if not isinstance(name, six.string_types):
        raise OperationTypeError('Name must be a string')

    # mode = ensure_mode_int(mode)
    info = host.fact.windows_file(name)

    # Not a file?!
    if info is False:
        raise OperationError('{0} exists and is not a file'.format(name))

    # Doesn't exist & we want it
    if not assume_present and info is None and present:
        if create_remote_dir:
            yield _create_remote_dir(state, host, name, user, group)

        yield 'New-Item -ItemType file {0}'.format(name)

#        if mode:
#            yield chmod(name, mode)
#        if user or group:
#            yield chown(name, user, group)

    # It exists and we don't want it
    elif (assume_present or info) and not present:
        yield 'Remove-Item {0}'.format(name)

#    # It exists & we want to ensure its state
#    elif (assume_present or info) and present:
#        if touch:
#            yield 'New-Item -ItemType file {0}'.format(name)
#
#        # Check mode
#        if mode and (not info or info['mode'] != mode):
#            yield chmod(name, mode)
#
#        # Check user/group
#        if (
#            (not info and (user or group))
#            or (user and info['user'] != user)
#            or (group and info['group'] != group)
#        ):
#            yield chown(name, user, group)


def windows_file(*args, **kwargs):
    # COMPAT
    # TODO: remove this
    logger.warning((
        'Use of `windows_files.windows_file` is deprecated, '
        'please use `windows_files.file` instead.'
    ))
    return file(*args, **kwargs)


def _create_remote_dir(state, host, remote_filename, user, group):
    # Always use POSIX style path as local might be Windows, remote always *nix
    remote_dirname = ntpath.dirname(remote_filename)
    if remote_dirname:
        yield directory(
            state, host, remote_dirname,
            user=user, group=group,
        )


@operation(pipeline_facts={
    'windows_directory': 'name',
})
def directory(
    name,
    present=True, assume_present=False,
    user=None, group=None, mode=None, recursive=False,
    state=None, host=None,
):
    '''
    Add/remove/update directories.

    + name: name/path of the remote folder
    + present: whether the folder should exist
    + assume_present: whether to assume the directory exists
    + TODO: user: user to own the folder
    + TODO: group: group to own the folder
    + TODO: mode: permissions of the folder
    + TODO: recursive: recursively apply user/group/mode

    Examples:

    .. code:: python

        files.directory(
            name='Ensure the c:\\temp\\dir_that_we_want_removed is removed',
            path='c:\\temp\\dir_that_we_want_removed',
            present=False,
        )

        files.directory(
            name='Ensure c:\\temp\\foo\\foo_dir exists',
            path='c:\\temp\\foo\\foo_dir',
            recursive=True,
        )

        # multiple directories
        dirs = ['c:\\temp\\foo_dir1', 'c:\\temp\\foo_dir2']
        for dir in dirs:
            files.directory(
                name='Ensure the directory `{}` exists'.format(dir),
                path=dir,
            )

    '''

    if not isinstance(name, six.string_types):
        raise OperationTypeError('Name must be a string')

    info = host.fact.windows_directory(name)

    # Not a directory?!
    if info is False:
        raise OperationError('{0} exists and is not a directory'.format(name))

    # Doesn't exist & we want it
    if not assume_present and info is None and present:
        yield 'mkdir -p {0}'.format(name)
#        if mode:
#            yield chmod(name, mode, recursive=recursive)
#        if user or group:
#            yield chown(name, user, group, recursive=recursive)
#
    # It exists and we don't want it
    elif (assume_present or info) and not present:
        # TODO: how to ensure we use 'ps'?
        # remove anything in the directory
        yield 'Get-ChildItem {0} -Recurse | Remove-Item'.format(name)
        # remove directory
        yield 'Remove-Item {0}'.format(name)

    # It exists & we want to ensure its state
#    elif (assume_present or info) and present:
#        # Check mode
#        if mode and (not info or info['mode'] != mode):
#            yield chmod(name, mode, recursive=recursive)
#
#        # Check user/group
#        if (
#            (not info and (user or group))
#            or (user and info['user'] != user)
#            or (group and info['group'] != group)
#        ):
#            yield chown(name, user, group, recursive=recursive)


def windows_directory(*args, **kwargs):
    # COMPAT
    # TODO: remove this
    logger.warning((
        'Use of `windows_files.windows_directory` is deprecated, '
        'please use `windows_files.directory` instead.'
    ))
    return directory(*args, **kwargs)



import json
import os

from tempfile import mkstemp

import click
import six

from pyinfra import local, logger
from pyinfra.api import QuoteString, StringCommand
from pyinfra.api.exceptions import ConnectError, InventoryError, PyinfraError
from pyinfra.api.util import get_file_io
from pyinfra.progress import progress_spinner

from .local import run_shell_command as run_local_shell_command
from .util import make_unix_command


def make_names_data(image=None):
    if not image:
        raise InventoryError('No docker base image provided!')

    # Save the image as the hostname, the image as data, @docker group
    yield '@docker/{0}'.format(image), {'docker_image': image}, ['@docker']


def _find_start_docker_container(container_id):
    docker_info = local.shell('docker container inspect {0}'.format(container_id))
    docker_info = json.loads(docker_info)[0]
    if docker_info['State']['Running'] is False:
        logger.info('Starting stopped container: {0}'.format(container_id))
        local.shell('docker container start {0}'.format(container_id))


def _start_docker_image(image_name):
    try:
        return local.shell(
            'docker run -d {0} tail -f /dev/null'.format(image_name),
            splitlines=True,
        )[-1]  # last line is the container ID
    except PyinfraError as e:
        raise ConnectError(e.args[0])


def connect(state, host):
    if 'docker_container_id' in host.host_data:  # user can provide a docker_container_id
        host.host_data['docker_container_no_disconnect'] = True
        return True

    with progress_spinner({'prepare docker container'}):
        try:
            # Check if the provided @docker/X is an existing container ID
            _find_start_docker_container(host.data.docker_image)
        except PyinfraError:
            container_id = _start_docker_image(host.data.docker_image)
        else:
            container_id = host.data.docker_image
            host.host_data['docker_container_no_disconnect'] = True

    host.host_data['docker_container_id'] = container_id
    return True


def disconnect(state, host):
    container_id = host.host_data['docker_container_id']

    if host.host_data.get('docker_container_no_disconnect'):
        logger.info('{0}docker build complete, container left running: {1}'.format(
            host.print_prefix, click.style(container_id, bold=True),
        ))
        return

    with progress_spinner({'docker commit'}):
        image_id = local.shell(
            'docker commit {0}'.format(container_id),
            splitlines=True,
        )[-1][7:19]  # last line is the image ID, get sha256:[XXXXXXXXXX]...

    with progress_spinner({'docker rm'}):
        local.shell(
            'docker rm -f {0}'.format(container_id),
        )

    logger.info('{0}docker build complete, image ID: {1}'.format(
        host.print_prefix, click.style(image_id, bold=True),
    ))


def run_shell_command(
    state, host, command,
    get_pty=False,
    timeout=None,
    stdin=None,
    success_exit_codes=None,
    print_output=False,
    print_input=False,
    return_combined_output=False,
    **command_kwargs
):
    container_id = host.host_data['docker_container_id']

    # Don't sudo/su in Docker - is this the right thing to do? Makes deploys that
    # target SSH systems work w/Docker out of the box (ie most docker commands
    # are run as root).
    for key in ('sudo', 'su_user'):
        command_kwargs.pop(key, None)

    command = make_unix_command(command, **command_kwargs)
    command = QuoteString(command)

    docker_flags = '-it' if get_pty else '-i'
    docker_command = StringCommand(
        'docker', 'exec', docker_flags, container_id,
        'sh', '-c', command,
    )

    return run_local_shell_command(
        state, host, docker_command,
        timeout=timeout,
        stdin=stdin,
        success_exit_codes=success_exit_codes,
        print_output=print_output,
        print_input=print_input,
        return_combined_output=return_combined_output,
    )


def put_file(
    state, host, filename_or_io, remote_filename,
    print_output=False, print_input=False,
    **kwargs  # ignored (sudo/etc)
):
    '''
    Upload a file/IO object to the target Docker container by copying it to a
    temporary location and then uploading it into the container using ``docker cp``.
    '''

    fd, temp_filename = mkstemp()

    try:
        # Load our file or IO object and write it to the temporary file
        with get_file_io(filename_or_io) as file_io:
            with open(temp_filename, 'wb') as temp_f:
                data = file_io.read()

                if isinstance(data, six.text_type):
                    data = data.encode()

                temp_f.write(data)

        docker_id = host.host_data['docker_container_id']
        docker_command = 'docker cp {0} {1}:{2}'.format(
            temp_filename,
            docker_id,
            remote_filename,
        )

        status, _, stderr = run_local_shell_command(
            state, host, docker_command,
            print_output=print_output,
            print_input=print_input,
        )
    finally:
        os.close(fd)
        os.remove(temp_filename)

    if not status:
        raise IOError('\n'.join(stderr))

    if print_output:
        click.echo('{0}file uploaded to container: {1}'.format(
            host.print_prefix, remote_filename,
        ), err=True)

    return status


def get_file(
    state, host, remote_filename, filename_or_io,
    print_output=False, print_input=False,
    **kwargs  # ignored (sudo/etc)
):
    '''
    Download a file from the target Docker container by copying it to a temporary
    location and then reading that into our final file/IO object.
    '''

    fd, temp_filename = mkstemp()

    try:
        docker_id = host.host_data['docker_container_id']
        docker_command = 'docker cp {0}:{1} {2}'.format(
            docker_id,
            remote_filename,
            temp_filename,
        )

        status, _, stderr = run_local_shell_command(
            state, host, docker_command,
            print_output=print_output,
            print_input=print_input,
        )

        # Load the temporary file and write it to our file or IO object
        with open(temp_filename) as temp_f:
            with get_file_io(filename_or_io, 'wb') as file_io:
                data = temp_f.read()

                if isinstance(data, six.text_type):
                    data = data.encode()

                file_io.write(data)
    finally:
        os.close(fd)
        os.remove(temp_filename)

    if not status:
        raise IOError('\n'.join(stderr))

    if print_output:
        click.echo('{0}file downloaded from container: {1}'.format(
            host.print_prefix, remote_filename,
        ), err=True)

    return status


EXECUTION_CONNECTOR = True



'''
This file as originally part of the "sshuserclient" pypi package. The GitHub
source has now vanished (https://github.com/tobald/sshuserclient).
'''

import glob
import os
import re

from paramiko import SSHConfig as ParamikoSSHConfig


class SSHConfig(ParamikoSSHConfig):
    '''
    an SSHConfig that supports includes directives
    https://github.com/paramiko/paramiko/pull/1194
    '''

    SETTINGS_REGEX = re.compile(r'(\w+)(?:\s*=\s*|\s+)(.+)')

    def parse(self, file_obj, parsed_files=None):
        '''
        Read an OpenSSH config from the given file object.

        :param file_obj: a file-like object to read the config file from
        '''
        host = {'host': ['*'], 'config': {}}
        for line in file_obj:
            # Strip any leading or trailing whitespace from the line.
            # Refer to https://github.com/paramiko/paramiko/issues/499
            line = line.strip()
            if not line or line.startswith('#'):
                continue

            match = re.match(self.SETTINGS_REGEX, line)
            if not match:
                raise Exception('Unparsable line {}'.format(line))
            key = match.group(1).lower()
            value = match.group(2)

            if key == 'host':
                self._config.append(host)
                host = {
                    'host': self._get_hosts(value),
                    'config': {},
                }
            elif key == 'proxycommand' and value.lower() == 'none':
                # Store 'none' as None; prior to 3.x, it will get stripped out
                # at the end (for compatibility with issue #415). After 3.x, it
                # will simply not get stripped, leaving a nice explicit marker.
                host['config'][key] = None
            elif key == 'include':
                # support for Include directive in ssh_config
                path = value
                # the path can be relative to its parent configuration file
                if os.path.isabs(path) is False and path[0] != '~':
                    folder = os.path.dirname(file_obj.name)
                    path = os.path.join(folder, path)

                # expand the user home path
                path = os.path.expanduser(path)
                if parsed_files is None:
                    parsed_files = []

                # parse every included file
                for filename in glob.iglob(path):
                    if os.path.isfile(filename):
                        if filename in parsed_files:
                            raise Exception(
                                'Include loop detected in ssh config file: %s' % filename,
                            )
                        with open(filename) as fd:
                            parsed_files.append(filename)
                            self.parse(fd, parsed_files)

            else:
                if value.startswith('"') and value.endswith('"'):
                    value = value[1:-1]

                # identityfile, localforward, remoteforward keys are special
                # cases, since they are allowed to be specified multiple times
                # and they should be tried in order of specification.
                if key in ['identityfile', 'localforward', 'remoteforward']:
                    if key in host['config']:
                        host['config'][key].append(value)
                    else:
                        host['config'][key] = [value]
                elif key not in host['config']:
                    host['config'][key] = value
        self._config.append(host)



import json

from unittest import TestCase

from mock import mock_open, patch

from pyinfra.api.connectors.mech import get_mech_options, make_names_data
from pyinfra.api.exceptions import InventoryError

FAKE_MECH_OPTIONS = {
    'groups': {
        'ubuntu16': ['mygroup'],
    },
    'data': {
        'centos7': {
            'somedata': 'somevalue',
        },
    },
}
FAKE_MECH_OPTIONS_DATA = json.dumps(FAKE_MECH_OPTIONS)


def fake_mech_shell(command, splitlines=None):
    if command == 'mech ls':
        return [
            'NAME	        ADDRESS	                                BOX	     VERSION	PATH',  # noqa: E501
            'ubuntu16	       192.168.2.226             bento/ubuntu-16.04	 201912.04.0	/Users/bob/somedir/ubuntu16',  # noqa: E501
            'centos7	       192.168.2.227                 bento/centos-7	 201912.05.0	/Users/bob/somedir/centos7',  # noqa: E501
            'centos6	       poweroff                      bento/centos-6	 201912.04.0	/Users/bob/somedir/centos6',  # noqa: E501
            'fedora31	               	                    bento/fedora-31	 201912.04.0	/Users/bob/somedir/fedora31',  # noqa: E501
        ]
    elif command == 'mech ssh-config ubuntu16':
        return [
            'Host ubuntu16',
            '  User vagrant',
            '  Port 22',
            '  UserKnownHostsFile /dev/null',
            '  StrictHostKeyChecking no',
            '  PasswordAuthentication no',
            '  IdentityFile path/to/key',
            '  IdentitiesOnly yes',
            '  LogLevel FATAL',
            '  HostName 192.168.2.226',
            '',
        ]
    elif command == 'mech ssh-config centos7':
        return [
            'Host centos7',
            '  User vagrant',
            '  Port 22',
            '  UserKnownHostsFile /dev/null',
            '  StrictHostKeyChecking no',
            '  PasswordAuthentication no',
            '  IdentityFile path/to/key',
            '  IdentitiesOnly yes',
            '  LogLevel FATAL',
            '  HostName 192.168.2.227',
            '',
        ]
    elif command == 'mech ssh-config centos6':
        return [
            'ERROR: Error: The virtual machine is not powered on: /Users/bob/debian8/.mech/debian-8.11-amd64.vmx',  # noqa: E501
            'This Mech machine is reporting that it is not yet ready for SSH. Make',
            'sure your machine is created and running and try again. Additionally,',
            'check the output of `mech status` to verify that the machine is in the',
            'state that you expect.',
        ]

    return []


@patch('pyinfra.api.connectors.mech.local.shell', fake_mech_shell)
class TestMechConnector(TestCase):
    def tearDown(self):
        get_mech_options.cache = {}

    @patch(
        'pyinfra.api.connectors.mech.open',
        mock_open(read_data=FAKE_MECH_OPTIONS_DATA),
        create=True,
    )
    @patch('pyinfra.api.connectors.mech.path.exists', lambda path: True)
    def test_make_names_data_with_options(self):
        data = make_names_data()

        assert data == [
            (
                '@mech/ubuntu16',
                {
                    'ssh_port': '22',
                    'ssh_user': 'vagrant',
                    'ssh_hostname': '192.168.2.226',
                    'ssh_key': 'path/to/key',
                },
                ['mygroup', '@mech'],
            ), (
                '@mech/centos7',
                {
                    'ssh_port': '22',
                    'ssh_user': 'vagrant',
                    'ssh_hostname': '192.168.2.227',
                    'ssh_key': 'path/to/key',
                    'somedata': 'somevalue',
                },
                ['@mech'],
            ),
        ]

    def test_make_names_data_with_limit(self):
        data = make_names_data(limit=('ubuntu16',))

        assert data == [
            (
                '@mech/ubuntu16',
                {
                    'ssh_port': '22',
                    'ssh_user': 'vagrant',
                    'ssh_hostname': '192.168.2.226',
                    'ssh_key': 'path/to/key',
                },
                ['@mech'],
            ),
        ]

    def test_make_names_data_no_matches(self):
        with self.assertRaises(InventoryError):
            make_names_data(limit='nope')



from .util import memoize


auth_kwargs = {
    'sudo': {
        'description': 'Execute/apply any changes with ``sudo``.',
        'default': lambda config: config.SUDO,
    },
    'sudo_user': {
        'description': 'Execute/apply any changes with ``sudo`` as a non-root user.',
        'default': lambda config: config.SUDO_USER,
    },
    'use_sudo_login': {
        'description': 'Execute ``sudo`` with a login shell.',
        'default': lambda config: config.USE_SUDO_LOGIN,
    },
    'use_sudo_password': {
        'description': 'Whether to use a password with ``sudo`` (will ask).',
        'default': lambda config: config.USE_SUDO_PASSWORD,
    },
    'preserve_sudo_env': {
        'description': 'Preserve the shell environment when using ``sudo``.',
        'default': lambda config: config.PRESERVE_SUDO_ENV,
    },
    'su_user': {
        'description': 'Execute/apply any changes with ``su``.',
        'default': lambda config: config.SU_USER,
    },
    'use_su_login': {
        'description': 'Execute ``su`` with a login shell.',
        'default': lambda config: config.USE_SU_LOGIN,
    },
    'preserve_su_env': {
        'description': 'Preserve the shell environment when using ``su``.',
        'default': lambda config: config.PRESERVE_SU_ENV,
    },
    'su_shell': {
        'description': (
            'Use this shell (instead of user login shell) when using ``su``). '
            'Only available under Linux, for use when using `su` with a user that '
            'has nologin/similar as their login shell.'
        ),
        'default': lambda config: config.SU_SHELL,
    },
}


def generate_env(config, value):
    env = config.ENV.copy()
    if value:
        env.update(value)
    return env


operation_kwargs = {
    'name': {
        'description': 'Name of the operation.',
    },
    'shell_executable': {
        'description': 'The shell to use. Defaults to ``sh`` (Unix) or ``cmd`` (Windows).',
        'default': lambda config: config.SHELL,
    },
    'chdir': {
        'description': 'Directory to switch to before executing the command.',
    },
    'env': {
        'description': 'Dictionary of environment variables to set.',
        'handler': generate_env,
    },
    'ignore_errors': {
        'description': 'Ignore errors when executing the operation.',
        'default': lambda config: config.IGNORE_ERRORS,
    },
    'success_exit_codes': {
        'description': 'List of exit codes to consider a success.',
        'default': lambda config: [0],
    },
    'timeout': 'Timeout for *each* command executed during the operation.',
    'get_pty': 'Whether to get a pseudoTTY when executing any commands.',
    'stdin': 'String or buffer to send to the stdin of any commands.',
    'precondition': 'Command to execute & check before the operation commands begin.',
    'postcondition': 'Command to execute & check after the operation commands complete.',
}

callback_kwargs = {
    'on_success': 'Callback function to execute on success.',
    'on_error': 'Callback function to execute on error.',
}

# Execution kwargs are global - ie must be identical for every host
execution_kwargs = {
    'parallel': {
        'description': 'Run this operation in batches of hosts.',
        'default': lambda config: config.PARALLEL,
    },
    'run_once': {
        'description': 'Only execute this operation once, on the first host to see it.',
        'default': lambda config: False,
    },
    'serial': {
        'description': 'Run this operation host by host, rather than in parallel.',
        'default': lambda config: False,
    },
}

OPERATION_KWARGS = {
    'Privilege & user escalation': auth_kwargs,
    'Operation control': operation_kwargs,
    'Callbacks': callback_kwargs,
    'Execution strategy (must be the same for all hosts)': execution_kwargs,
}


def get_execution_kwarg_keys():
    return list(execution_kwargs.keys())


@memoize
def get_executor_kwarg_keys():
    keys = set()
    keys.update(auth_kwargs.keys(), operation_kwargs.keys())
    keys.difference_update({'name', 'ignore_errors', 'precondition', 'postcondition'})
    return list(keys)


def pop_global_op_kwargs(state, kwargs):
    '''
    Pop and return operation global keyword arguments.
    '''

    meta_kwargs = state.deploy_kwargs or {}

    def get_kwarg(key, default=None):
        has_key = key in kwargs
        value = kwargs.pop(key, meta_kwargs.get(key, default))
        return value, has_key

    global_kwargs = {}
    global_kwarg_keys = []

    for _, kwarg_configs in OPERATION_KWARGS.items():
        for key, config in kwarg_configs.items():
            handler = None
            default = None

            if isinstance(config, dict):
                handler = config.get('handler')
                default = config.get('default')
                if default:
                    default = default(state.config)

            value, has_key = get_kwarg(key, default=default)
            if handler:
                value = handler(state.config, value)

            if has_key:
                global_kwarg_keys.append(key)

            global_kwargs[key] = value

    return global_kwargs, global_kwarg_keys



from __future__ import print_function

import json
import warnings

from os import listdir, path
from unittest import TestCase

import six

from pyinfra.api import StringCommand
from pyinfra.api.facts import FACTS, ShortFactBase
from pyinfra_cli.util import json_encode

from .util import get_command_string, JsonTest

# show full diff on json
TestCase.maxDiff = None


def _make_command(command_attribute, args):
    if callable(command_attribute):
        if not isinstance(args, list):
            args = [args]

        return command_attribute(*args)
    return command_attribute


def make_fact_tests(fact_name):
    fact = FACTS[fact_name]()

    @six.add_metaclass(JsonTest)
    class TestTests(TestCase):
        jsontest_files = path.join('tests', 'facts', fact_name)
        jsontest_prefix = 'test_{0}_'.format(fact_name)

        def jsontest_function(self, test_name, test_data, fact=fact):
            short_fact = None

            if isinstance(fact, ShortFactBase):
                short_fact = fact
                fact = fact.fact()

            test_args = test_data.get('arg', [])
            command = _make_command(fact.command, test_args)

            if 'command' in test_data:
                assert get_command_string(StringCommand(command)) == test_data['command']
            else:
                warnings.warn('No command set for test: {0} (got "{1}")'.format(
                    test_name, command,
                ))

            requires_command = _make_command(fact.requires_command, test_args)

            if requires_command:
                if 'requires_command' in test_data:
                    assert requires_command == test_data['requires_command']
                else:
                    warnings.warn('No requires command set for test: {0} (got "{1}")'.format(
                        test_name, requires_command,
                    ))

            data = fact.process(test_data['output'])
            if short_fact:
                data = short_fact.process_data(data)

            # Encode/decode data to ensure datetimes/etc become JSON
            data = json.loads(json.dumps(data, default=json_encode))
            try:
                assert data == test_data['fact']
            except AssertionError as e:
                print()
                print('--> GOT:\n', json.dumps(data, indent=4, default=json_encode))
                print('--> WANT:', json.dumps(
                    test_data['fact'], indent=4, default=json_encode,
                ))
                raise e

    TestTests.__name__ = 'Fact{0}'.format(fact_name)
    return TestTests


# Find available fact tests
fact_tests = sorted([
    filename
    for filename in listdir(path.join('tests', 'facts'))
    if path.isdir(path.join('tests', 'facts', filename))
])

# Generate the classes, attaching to local
for fact_name in fact_tests:
    locals()[fact_name] = make_fact_tests(fact_name)



'''
Deploys come in two forms: on-disk, eg deploy.py, and @deploy wrapped functions.
The latter enable re-usable (across CLI and API based execution) pyinfra extension
creation (eg pyinfra-openstack).
'''

from functools import wraps

import six

import pyinfra

from pyinfra import logger, pseudo_host, pseudo_state

from .exceptions import PyinfraError
from .host import Host
from .operation_kwargs import pop_global_op_kwargs
from .state import State
from .util import get_call_location, get_caller_frameinfo, memoize


@memoize
def show_state_host_arguments_warning(call_location):
    logger.warning((
        '{0}\n\tPassing `state` and `host` as the first two arguments to deploys is '
        'deprecated, please use `state` and `host` arguments.'
    ).format(call_location))


def add_deploy(state, deploy_func, *args, **kwargs):
    '''
    Prepare & add an deploy to pyinfra.state by executing it on all hosts.

    Args:
        state (``pyinfra.api.State`` obj): the deploy state to add the operation
        deploy_func (function): the operation function from one of the modules,
        ie ``server.user``
        args/kwargs: passed to the operation function
    '''

    if pyinfra.is_cli:
        raise PyinfraError((
            '`add_deploy` should not be called when pyinfra is executing in CLI mode! ({0})'
        ).format(get_call_location()))

    kwargs['state'] = state

    hosts = kwargs.pop('host', state.inventory.iter_active_hosts())
    if isinstance(hosts, Host):
        hosts = [hosts]

    # Append operations called in this deploy to the current order
    kwargs['_op_order_number'] = len(state.op_meta)

    for host in hosts:
        kwargs['host'] = host
        deploy_func(*args, **kwargs)


def deploy(func_or_name, data_defaults=None):
    '''
    Decorator that takes a deploy function (normally from a pyinfra_* package)
    and wraps any operations called inside with any deploy-wide kwargs/data.
    '''

    # If not decorating, return function with config attached
    if isinstance(func_or_name, six.string_types):
        name = func_or_name

        def decorator(f):
            setattr(f, 'deploy_name', name)

            if data_defaults:
                setattr(f, 'deploy_data', data_defaults)

            return deploy(f)

        return decorator

    # Actually decorate!
    func = func_or_name

    @wraps(func)
    def decorated_func(*args, **kwargs):
        # State & host passed in as kwargs (API, nested op, @deploy op)
        if 'state' in kwargs and 'host' in kwargs:
            state = kwargs['state']
            host = kwargs['host']

        # State & host passed in as first two arguments (LEGACY)
        elif len(args) >= 2 and isinstance(args[0], State) and isinstance(args[1], Host):
            show_state_host_arguments_warning(get_call_location())
            state = kwargs['state'] = args[0]
            host = kwargs['host'] = args[1]
            args_copy = list(args)
            args = args_copy[2:]

        # Finally, still no state+host? Use pseudo if we're CLI mode, or fail
        elif pyinfra.is_cli:
            state = kwargs['state'] = pseudo_state._module
            host = kwargs['host'] = pseudo_host._module

            if not state or not host or state.in_deploy:
                raise PyinfraError((
                    'Nested deploy called without state/host: {0} ({1})'
                ).format(func, get_call_location()))

        else:
            raise PyinfraError((
                'Deploy called without state/host: {0} ({1})'
            ).format(func, get_call_location()))

        deploy_kwargs, _ = pop_global_op_kwargs(state, kwargs)

        # Name the deploy
        deploy_name = getattr(func, 'deploy_name', func.__name__)
        deploy_data = getattr(func, 'deploy_data', None)

        deploy_op_order = None

        if not pyinfra.is_cli:
            # API mode: `add_deploy` provides the order number
            op_order_number = kwargs.pop('_op_order_number', None)
            if op_order_number is not None:
                deploy_op_order = [op_order_number]
            # API mode: nested deploy wrapped function call
            elif state.deploy_op_order:
                frameinfo = get_caller_frameinfo()
                deploy_op_order = state.deploy_op_order + [frameinfo.lineno]
            else:
                raise PyinfraError((
                    'Operation order number not provided in API mode - '
                    'you must use `add_deploy` to add operations.'
                ))

        with state.deploy(
            name=deploy_name,
            kwargs=deploy_kwargs,
            data=deploy_data,
            deploy_op_order=deploy_op_order,
        ):
            func(*args, **kwargs)

    return decorated_func



'''
.. warning::

    This is a legacy module providing backwards compatibility. Please use the other modules:

    + :doc:`./systemd`
    + :doc:`./upstart`
    + :doc:`./launchd`
    + :doc:`./sysvinit`
    + :doc:`./bsdinit`
    + `server.service operation <server.html#server-service>`_
'''

from pyinfra import logger

from . import bsdinit, launchd, server, systemd, sysvinit, upstart


def _make_legacy_operation(legacy_op, op_func):
    def wrapper(*args, **kwargs):
        logger.warning((
            'The `init.{0}` operation is deprecated, '
            'please us the `{1}.{2}` operation.'
        ).format(
            legacy_op,
            op_func.__module__.replace('pyinfra.operations.', ''),
            op_func.__name__,
        ))
        return op_func(*args, **kwargs)

    wrapper._pyinfra_op = op_func
    return wrapper


d = _make_legacy_operation('d', sysvinit.service)
d_enable = _make_legacy_operation('d_enable', sysvinit.enable)
rc = _make_legacy_operation('rc', bsdinit.service)
upstart = _make_legacy_operation('upstart', upstart.service)
systemd = _make_legacy_operation('systemd', systemd.service)
launchd = _make_legacy_operation('launchd', launchd.service)
service = _make_legacy_operation('service', server.service)



from __future__ import division

import math
import os
import platform
import sys

from collections import deque
from contextlib import contextmanager
from threading import Event, Thread
from time import sleep

import pyinfra

IS_WINDOWS = platform.system() == 'Windows'

WAIT_TIME = 1 / 5
WAIT_CHARS = deque(('-', '/', '|', '\\'))

# Hacky way of getting terminal size (so can clear lines)
# Source: http://stackoverflow.com/questions/566746
IS_TTY = sys.stdout.isatty() and sys.stderr.isatty()
TERMINAL_WIDTH = 0

if IS_TTY:
    try:
        TERMINAL_WIDTH = os.get_terminal_size().columns
    except AttributeError:
        if not IS_WINDOWS:
            terminal_size = os.popen('stty size', 'r').read().split()
            if len(terminal_size) == 2:
                TERMINAL_WIDTH = int(terminal_size[1])


def _print_spinner(stop_event, progress_queue):
    if not IS_TTY or os.environ.get('PYINFRA_PROGRESS') == 'off':
        return

    progress = ''
    text = ''

    while True:
        # Stop when asked too
        if stop_event.is_set():
            break

        WAIT_CHARS.rotate(1)

        try:
            progress = progress_queue[-1]
        except IndexError:
            pass

        text = '    {0}'.format(
            ' '.join((WAIT_CHARS[0], progress)),
        )
        text = '{0}\r'.format(text)

        sys.stderr.write(text)
        sys.stderr.flush()

        # In pyinfra_cli's __main__ we set stdout & stderr to be line buffered,
        # so write this escape code (clear line) into the buffer but don't flush,
        # such that any next print/log/etc clear the line first.
        if not IS_WINDOWS:
            sys.stderr.write('\033[K')

        sleep(WAIT_TIME)


@contextmanager
def progress_spinner(items, prefix_message=None):
    # If there's no pseudo state we're not in CLI mode, so just return a noop
    # handler and exit.
    if not pyinfra.is_cli:
        yield lambda complete_item: None
        return

    if not isinstance(items, set):
        items = set(items)

    total_items = len(items)
    stop_event = Event()

    def make_progress_message(include_items=True):
        message_bits = []

        # If we only have 1 item, don't show %
        if total_items > 1:
            percentage_complete = 0
            complete = total_items - len(items)
            percentage_complete = int(math.floor(complete / total_items * 100))
            message_bits.append('{0}% ({1}/{2})'.format(
                percentage_complete,
                complete,
                total_items,
            ))

        if prefix_message:
            message_bits.append(prefix_message)

        if include_items and items:
            # Plus 3 for the " - " joining below
            message_length = sum((len(message) + 3) for message in message_bits)
            # -8 for padding left+right, -2 for {} wrapping
            items_allowed_width = TERMINAL_WIDTH - 10 - message_length

            if items_allowed_width > 0:
                items_string = '{%s}' % (', '.join('{0}'.format(i) for i in items))
                if len(items_string) >= items_allowed_width:
                    items_string = '%s...}' % (
                        # -3 for the ...
                        items_string[:items_allowed_width - 3],
                    )

                message_bits.append(items_string)

        return ' - '.join(message_bits)

    progress_queue = deque((make_progress_message(),))

    def progress(complete_item):
        if complete_item not in items:
            raise ValueError('Invalid complete item: {0} not in {1}'.format(
                complete_item, items,
            ))

        items.remove(complete_item)
        progress_queue.append(make_progress_message())

    # Kick off the spinner thread
    spinner_thread = Thread(
        target=_print_spinner,
        args=(stop_event, progress_queue),
    )
    spinner_thread.daemon = True
    spinner_thread.start()

    # Yield allowing the actual code the spinner waits for to run
    yield progress

    # Finally, stop the spinner
    stop_event.set()
    spinner_thread.join()



'''
The server module takes care of os-level state. Targets POSIX compatibility, tested on
Linux/BSD.
'''

from __future__ import division, unicode_literals

from itertools import tee
from os import path
from time import sleep

import six

from six.moves import filterfalse, shlex_quote

from pyinfra.api import FunctionCommand, operation, OperationError, StringCommand
from pyinfra.api.util import try_int

from . import bsdinit, files, systemd, sysvinit, upstart
from .util.files import chmod, sed_replace


@operation
def reboot(delay=10, interval=1, reboot_timeout=300, state=None, host=None):
    '''
    Reboot the server and wait for reconnection.

    + delay: number of seconds to wait before attempting reconnect
    + interval: interval (s) between reconnect attempts
    + reboot_timeout: total time before giving up reconnecting

    Example:

    .. code:: python

        server.reboot(
            name='Reboot the server and wait to reconnect',
            delay=60,
            reboot_timeout=600,
        )
    '''

    yield StringCommand('reboot', success_exit_codes=[0, -1])  # -1 being error/disconnected

    def wait_and_reconnect(state, host):  # pragma: no cover
        sleep(delay)
        max_retries = round(reboot_timeout / interval)

        host.connection = None  # remove the connection object
        retries = 0

        while True:
            host.connect(show_errors=False)
            if host.connection:
                break

            if retries > max_retries:
                raise Exception((
                    'Server did not reboot in time (reboot_timeout={0}s)'
                ).format(reboot_timeout))

            sleep(interval)
            retries += 1

    yield FunctionCommand(wait_and_reconnect, (), {})


@operation
def wait(port=None, state=None, host=None):
    '''
    Waits for a port to come active on the target machine. Requires netstat, checks every
    second.

    + port: port number to wait for

    Example:

    .. code:: python

        server.wait(
            name='Wait for webserver to start',
            port=80,
        )
    '''

    yield r'''
        while ! (netstat -an | grep LISTEN | grep -e "\.{0}" -e ":{0}"); do
            echo "waiting for port {0}..."
            sleep 1
        done
    '''.format(port)


@operation
def shell(commands, state=None, host=None):
    '''
    Run raw shell code on server during a deploy. If the command would
    modify data that would be in a fact, the fact would not be updated
    since facts are only run at the start of a deploy.

    + commands: command or list of commands to execute on the remote server

    Example:

    .. code:: python

        server.shell(
            name='Run lxd auto init',
            commands=['lxd init --auto'],
        )
    '''

    # Ensure we have a list
    if isinstance(commands, six.string_types):
        commands = [commands]

    for command in commands:
        yield command


@operation
def script(src, state=None, host=None):
    '''
    Upload and execute a local script on the remote host.

    + src: local script filename to upload & execute

    Example:

    .. code:: python

        # Note: This assumes there is a file in files/hello.bash locally.
        server.script(
            name='Hello',
            src='files/hello.bash',
        )
    '''

    temp_file = state.get_temp_filename(src)
    yield files.put(src, temp_file, state=state, host=host)

    yield chmod(temp_file, '+x')
    yield temp_file


@operation
def script_template(src, state=None, host=None, **data):
    '''
    Generate, upload and execute a local script template on the remote host.

    + src: local script template filename

    Example:

    .. code:: python

        # Example showing how to pass python variable to a script template file.
        # The .j2 file can use `{{ some_var }}` to be interpolated.
        # To see output need to run pyinfra with '-v'
        # Note: This assumes there is a file in templates/hello2.bash.j2 locally.
        some_var = 'blah blah blah '
        server.script_template(
            name='Hello from script',
            src='templates/hello2.bash.j2',
            some_var=some_var,
        )
    '''

    temp_file = state.get_temp_filename('{0}{1}'.format(src, data))
    yield files.template(src, temp_file, state=state, host=host, **data)

    yield chmod(temp_file, '+x')
    yield temp_file


@operation
def modprobe(module, present=True, force=False, state=None, host=None):
    '''
    Load/unload kernel modules.

    + module: name of the module to manage
    + present: whether the module should be loaded or not
    + force: whether to force any add/remove modules

    Example:

    .. code:: python

        server.modprobe(
            name='Silly example for modprobe',
            module='floppy',
        )
    '''
    list_value = (
        [module]
        if isinstance(module, six.string_types)
        else module
    )

    # NOTE: https://docs.python.org/3/library/itertools.html#itertools-recipes
    def partition(predicate, iterable):
        t1, t2 = tee(iterable)
        return list(filter(predicate, t2)), list(filterfalse(predicate, t1))

    modules = host.fact.kernel_modules
    present_mods, missing_mods = partition(lambda mod: mod in modules, list_value)

    args = ''
    if force:
        args = ' -f'

    # Module is loaded and we don't want it?
    if not present and present_mods:
        yield 'modprobe{0} -r -a {1}'.format(args, ' '.join(present_mods))

    # Module isn't loaded and we want it?
    elif present and missing_mods:
        yield 'modprobe{0} -a {1}'.format(args, ' '.join(missing_mods))

    else:
        host.noop('{0} {1} {2} {3}'.format(
            'modules' if len(list_value) > 1 else 'module',
            '/'.join(list_value),
            'are' if len(list_value) > 1 else 'is',
            'loaded' if present else 'not loaded',
        ))


@operation
def mount(
    path,
    mounted=True, options=None,
    # TODO: do we want to manage fstab here?
    # update_fstab=False, device=None, fs_type=None,
    state=None, host=None,
):
    '''
    Manage mounted filesystems.

    + path: the path of the mounted filesystem
    + mounted: whether the filesystem should be mounted
    + options: the mount options

    Options:
        If the currently mounted filesystem does not have all of the provided
        options it will be remounted with the options provided.

    ``/etc/fstab``:
        This operation does not attempt to modify the on disk fstab file - for
        that you should use the `files.line operation <./files.html#files-line>`_.
    '''

    options = options or []
    options_string = ','.join(options)

    mounts = host.fact.mounts
    is_mounted = path in mounts

    # Want mount but don't have?
    if mounted and not is_mounted:
        yield 'mount{0} {1}'.format(
            ' -o {0}'.format(options_string) if options_string else '',
            path,
        )

    # Want no mount but mounted?
    elif mounted is False and is_mounted:
        yield 'umount {0}'.format(path)

    # Want mount and is mounted! Check the options
    elif is_mounted and mounted and options:
        mounted_options = mounts[path]['options']
        needed_options = set(options) - set(mounted_options)
        if needed_options:
            yield 'mount -o remount,{0} {1}'.format(options_string, path)

    else:
        host.noop('filesystem {0} is {1}'.format(
            path,
            'mounted' if mounted else 'not mounted',
        ))


@operation
def hostname(hostname, hostname_file=None, state=None, host=None):
    '''
    Set the system hostname using ``hostnamectl`` or ``hostname`` on older systems.

    + hostname: the hostname that should be set
    + hostname_file: the file that permanently sets the hostname

    Hostname file:
        The hostname file only matters no systems that do not have ``hostnamectl``,
        which is part of ``systemd``.

        By default pyinfra will auto detect this by targetting ``/etc/hostname``
        on Linux and ``/etc/myname`` on OpenBSD.

        To completely disable writing the hostname file, set ``hostname_file=False``.

    Example:

    .. code:: python

        server.hostname(
            name='Set the hostname',
            hostname='server1.example.com',
        )
    '''

    current_hostname = host.fact.hostname

    if host.fact.which('hostnamectl'):
        if current_hostname != hostname:
            yield 'hostnamectl set-hostname {0}'.format(hostname)
        else:
            host.noop('hostname is set')
        return

    if hostname_file is None:
        os = host.fact.os

        if os == 'Linux':
            hostname_file = '/etc/hostname'
        elif os == 'OpenBSD':
            hostname_file = '/etc/myname'

    if current_hostname != hostname:
        yield 'hostname {0}'.format(hostname)
    else:
        host.noop('hostname is set')

    if hostname_file:
        # Create a whole new hostname file
        file = six.StringIO('{0}\n'.format(hostname))

        # And ensure it exists
        yield files.put(
            file, hostname_file,
            state=state,
            host=host,
        )


@operation
def sysctl(
    key, value,
    persist=False, persist_file='/etc/sysctl.conf',
    state=None, host=None,
):
    '''
    Edit sysctl configuration.

    + key: name of the sysctl setting to ensure
    + value: the value or list of values the sysctl should be
    + persist: whether to write this sysctl to the config
    + persist_file: file to write the sysctl to persist on reboot

    Example:

    .. code:: python

        server.sysctl(
            name='Change the fs.file-max value',
            key='fs.file-max',
            value=100000,
            persist=True,
        )
    '''

    string_value = (
        ' '.join(['{0}'.format(v) for v in value])
        if isinstance(value, list)
        else value
    )

    value = (
        [try_int(v) for v in value]
        if isinstance(value, list)
        else try_int(value)
    )

    existing_value = host.fact.sysctl.get(key)
    if not existing_value or existing_value != value:
        yield "sysctl {0}='{1}'".format(key, string_value)
        host.fact.sysctl[key] = value
    else:
        host.noop('sysctl {0} is set to {1}'.format(key, string_value))

    if persist:
        yield files.line(
            path=persist_file,
            line='{0}[[:space:]]*=[[:space:]]*{1}'.format(key, string_value),
            replace='{0} = {1}'.format(key, string_value),
            state=state,
            host=host,
        )


@operation
def service(
    service,
    running=True, restarted=False, reloaded=False,
    command=None, enabled=None,
    state=None, host=None,
):
    '''
    Manage the state of services. This command checks for the presence of all the
    Linux init systems ``pyinfra`` can handle and executes the relevant operation.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + reloaded: whether the service should be reloaded
    + command: custom command execute
    + enabled: whether this service should be enabled/disabled on boot

    Example:

    .. code:: python

        server.service(
            name='Enable open-vm-tools service',
            service='open-vm-tools',
            enabled=True,
        )
    '''

    if host.fact.which('systemctl'):
        service_operation = systemd.service

    elif host.fact.which('initctl'):
        service_operation = upstart.service

    elif host.fact.directory('/etc/init.d'):
        service_operation = sysvinit.service

    elif host.fact.directory('/etc/rc.d'):
        service_operation = bsdinit.service

    else:
        raise OperationError((
            'No init system found '
            '(no systemctl, initctl, /etc/init.d or /etc/rc.d found)'
        ))

    yield service_operation(
        service,
        running=running, restarted=restarted, reloaded=reloaded,
        command=command, enabled=enabled,
        state=state, host=host,
    )


@operation
def crontab(
    command,
    present=True,
    user=None,
    cron_name=None,
    minute='*',
    hour='*',
    month='*',
    day_of_week='*',
    day_of_month='*',
    special_time=None,
    interpolate_variables=False,
    state=None, host=None,
):
    '''
    Add/remove/update crontab entries.

    + command: the command for the cron
    + present: whether this cron command should exist
    + user: the user whose crontab to manage
    + cron_name: name the cronjob so future changes to the command will overwrite
    + minute: which minutes to execute the cron
    + hour: which hours to execute the cron
    + month: which months to execute the cron
    + day_of_week: which day of the week to execute the cron
    + day_of_month: which day of the month to execute the cron
    + special_time: cron "nickname" time (@reboot, @daily, etc), overrides others
    + interpolate_variables: whether to interpolate variables in ``command``

    Cron commands:
        Unless ``name`` is specified the command is used to identify crontab entries.
        This means commands must be unique within a given users crontab. If you require
        multiple identical commands, provide a different name argument for each.

    Special times:
        When provided, ``special_time`` will be used instead of any values passed in
        for ``minute``/``hour``/``month``/``day_of_week``/``day_of_month``.

    Example:

    .. code:: python

        # simple example for a crontab
        server.crontab(
            name='Backup /etc weekly',
            command='/bin/tar cf /tmp/etc_bup.tar /etc',
            name='backup_etc',
            day_of_week=0,
            hour=1,
            minute=0,
        )
    '''

    def comma_sep(value):
        if isinstance(value, (list, tuple)):
            return ','.join('{0}'.format(v) for v in value)
        return value

    minute = comma_sep(minute)
    hour = comma_sep(hour)
    month = comma_sep(month)
    day_of_week = comma_sep(day_of_week)
    day_of_month = comma_sep(day_of_month)

    crontab = host.fact.crontab(user)
    name_comment = '# pyinfra-name={0}'.format(cron_name)

    existing_crontab = crontab.get(command)
    existing_crontab_command = command
    existing_crontab_match = command

    if not existing_crontab and cron_name:  # find the crontab by name if provided
        for cmd, details in crontab.items():
            if name_comment in details['comments']:
                existing_crontab = details
                existing_crontab_match = cmd
                existing_crontab_command = cmd

    exists = existing_crontab is not None

    edit_commands = []
    temp_filename = state.get_temp_filename()

    if special_time:
        new_crontab_line = '{0} {1}'.format(special_time, command)
    else:
        new_crontab_line = '{minute} {hour} {day_of_month} {month} {day_of_week} {command}'.format(
            minute=minute,
            hour=hour,
            day_of_month=day_of_month,
            month=month,
            day_of_week=day_of_week,
            command=command,
        )

    existing_crontab_match = '.*{0}.*'.format(existing_crontab_match)

    # Don't want the cron and it does exist? Remove the line
    if not present and exists:
        edit_commands.append(sed_replace(
            temp_filename, existing_crontab_match, '',
            interpolate_variables=interpolate_variables,
        ))

    # Want the cron but it doesn't exist? Append the line
    elif present and not exists:
        if cron_name:
            if crontab:  # append a blank line if cron entries already exist
                edit_commands.append("echo '' >> {0}".format(temp_filename))
            edit_commands.append('echo {0} >> {1}'.format(
                shlex_quote(name_comment), temp_filename,
            ))

        edit_commands.append('echo {0} >> {1}'.format(
            shlex_quote(new_crontab_line), temp_filename,
        ))

    # We have the cron and it exists, do it's details? If not, replace the line
    elif present and exists:
        if any((
            special_time != existing_crontab.get('special_time'),
            minute != existing_crontab.get('minute'),
            hour != existing_crontab.get('hour'),
            month != existing_crontab.get('month'),
            day_of_week != existing_crontab.get('day_of_week'),
            day_of_month != existing_crontab.get('day_of_month'),
            existing_crontab_command != command,
        )):
            edit_commands.append(sed_replace(
                temp_filename, existing_crontab_match, new_crontab_line,
                interpolate_variables=interpolate_variables,
            ))

    if edit_commands:
        crontab_args = []
        if user:
            crontab_args.append('-u {0}'.format(user))

        # List the crontab into a temporary file if it exists
        if crontab:
            yield 'crontab -l {0} > {1}'.format(' '.join(crontab_args), temp_filename)

        # Now yield any edits
        for edit_command in edit_commands:
            yield edit_command

        # Finally, use the tempfile to write a new crontab
        yield 'crontab {0} {1}'.format(' '.join(crontab_args), temp_filename)

        # Update the crontab fact
        if present:
            crontab[command] = {
                'special_time': special_time,
                'minute': minute,
                'hour': hour,
                'month': month,
                'day_of_week': day_of_week,
                'day_of_month': day_of_month,
                'comments': [cron_name] if cron_name else [],
            }
        else:
            crontab.pop(command)
    else:
        host.noop('crontab {0} {1}'.format(
            command,
            'exists' if present else 'does not exist',
        ))


@operation
def group(group, present=True, system=False, gid=None, state=None, host=None):
    '''
    Add/remove system groups.

    + group: name of the group to ensure
    + present: whether the group should be present or not
    + system: whether to create a system group

    System users:
        System users don't exist on BSD, so the argument is ignored for BSD targets.

    Examples:

    .. code:: python

        server.group(
            name='Create docker group',
            group='docker',
        )

        # multiple groups
        for group in ['wheel', 'lusers']:
            server.group(
                name=f'Create the group {group}',
                group=group,
            )
    '''

    groups = host.fact.groups
    is_present = group in groups

    # Group exists but we don't want them?
    if not present and is_present:
        yield 'groupdel {0}'.format(group)
        groups.remove(group)

    # Group doesn't exist and we want it?
    elif present and not is_present:
        args = []

        # BSD doesn't do system users
        if system and 'BSD' not in host.fact.os:
            args.append('-r')

        args.append(group)

        if gid:
            args.append('--gid {0}'.format(gid))

        # Groups are often added by other operations (package installs), so check
        # for the group at runtime before adding.
        yield "grep '{0}:' /etc/group || groupadd {1}".format(
            group,
            ' '.join(args),
        )
        groups.append(group)


@operation
def user(
    user,
    present=True, home=None, shell=None, group=None, groups=None,
    public_keys=None, delete_keys=False, ensure_home=True,
    system=False, uid=None, comment=None, add_deploy_dir=True,
    unique=True,
    state=None, host=None,
):
    '''
    Add/remove/update system users & their ssh `authorized_keys`.

    + user: name of the user to ensure
    + present: whether this user should exist
    + home: the users home directory
    + shell: the users shell
    + group: the users primary group
    + groups: the users secondary groups
    + public_keys: list of public keys to attach to this user, ``home`` must be specified
    + delete_keys: whether to remove any keys not specified in ``public_keys``
    + ensure_home: whether to ensure the ``home`` directory exists
    + system: whether to create a system account
    + comment: the user GECOS comment
    + add_deploy_dir: any public_key filenames are relative to the deploy directory
    + unique: prevent creating users with duplicate UID

    Home directory:
        When ``ensure_home`` or ``public_keys`` are provided, ``home`` defaults to
        ``/home/{name}``.

    Public keys:
        These can be provided as strings containing the public key or as a path to
        a public key file which ``pyinfra`` will read.

    Examples:

    .. code:: python

        server.user(
            name='Ensure user is removed',
            user='kevin',
            present=False,
        )

        server.user(
            name='Ensure myweb user exists',
            user='myweb',
            shell='/bin/bash',
        )

        # multiple users
        for user in ['kevin', 'bob']:
            server.user(
                name=f'Ensure user {user} is removed',
                user=user,
                present=False,
            )
    '''

    users = host.fact.users
    existing_user = users.get(user)

    if groups is None:
        groups = []

    if home is None:
        home = '/home/{0}'.format(user)

    # User not wanted?
    if not present:
        if existing_user:
            yield 'userdel {0}'.format(user)
            users.pop(user)
        return

    # User doesn't exist but we want them?
    if present and existing_user is None:
        # Create the user w/home/shell
        args = []

        if home:
            args.append('-d {0}'.format(home))

        if shell:
            args.append('-s {0}'.format(shell))

        if group:
            args.append('-g {0}'.format(group))

        if groups:
            args.append('-G {0}'.format(','.join(groups)))

        if system and 'BSD' not in host.fact.os:
            args.append('-r')

        if uid:
            args.append('--uid {0}'.format(uid))

        if comment:
            args.append("-c '{0}'".format(comment))

        if not unique:
            args.append('-o')

        # Users are often added by other operations (package installs), so check
        # for the user at runtime before adding.
        yield "grep '{1}:' /etc/passwd || useradd {0} {1}".format(
            ' '.join(args),
            user,
        )
        users[user] = {
            'comment': comment,
            'home': home,
            'shell': shell,
            'group': group,
            'groups': groups,
        }
        # Flag this username as "to be added" in the state
        state.add_will_add_user(user)

    # User exists and we want them, check home/shell/keys
    else:
        args = []

        # Check homedir
        if home and existing_user['home'] != home:
            args.append('-d {0}'.format(home))

        # Check shell
        if shell and existing_user['shell'] != shell:
            args.append('-s {0}'.format(shell))

        # Check primary group
        if group and existing_user['group'] != group:
            args.append('-g {0}'.format(group))

        # Check secondary groups, if defined
        if groups and set(existing_user['groups']) != set(groups):
            args.append('-G {0}'.format(','.join(groups)))

        if comment and existing_user['comment'] != comment:
            args.append("-c '{0}'".format(comment))

        # Need to mod the user?
        if args:
            yield 'usermod {0} {1}'.format(' '.join(args), user)
            if comment:
                existing_user['comment'] = comment
            if home:
                existing_user['home'] = home
            if shell:
                existing_user['shell'] = shell
            if group:
                existing_user['group'] = group
            if groups:
                existing_user['groups'] = groups

    # Ensure home directory ownership
    if ensure_home:
        yield files.directory(
            home,
            user=user, group=user,
            state=state, host=host,
        )

    # Add SSH keys
    if public_keys is not None:
        if isinstance(public_keys, six.string_types):
            public_keys = [public_keys]

        def read_any_pub_key_file(key):
            try_path = key
            if add_deploy_dir and state.deploy_dir:
                try_path = path.join(state.deploy_dir, key)

            if path.exists(try_path):
                with open(try_path, 'r') as f:
                    return f.read()

            return key

        public_keys = list(map(read_any_pub_key_file, public_keys))

        # Ensure .ssh directory
        # note that this always outputs commands unless the SSH user has access to the
        # authorized_keys file, ie the SSH user is the user defined in this function
        yield files.directory(
            '{0}/.ssh'.format(home),
            user=user,
            group=user,
            mode=700,
            state=state,
            host=host,
        )

        filename = '{0}/.ssh/authorized_keys'.format(home)

        if delete_keys:
            # Create a whole new authorized_keys file
            keys_file = six.StringIO('{0}\n'.format(
                '\n'.join(public_keys),
            ))

            # And ensure it exists
            yield files.put(
                src=keys_file,
                dest=filename,
                user=user,
                group=user,
                mode=600,
                state=state,
                host=host,
            )

        else:
            # Ensure authorized_keys exists
            yield files.file(
                path=filename,
                user=user,
                group=user,
                mode=600,
                state=state,
                host=host,
            )

            # And every public key is present
            for key in public_keys:
                yield files.line(path=filename, line=key, state=state, host=host)



'''
Manage apk packages.
'''

from pyinfra.api import operation

from .util.packaging import ensure_packages


@operation
def upgrade(state=None, host=None):
    '''
    Upgrades all apk packages.
    '''

    yield 'apk upgrade'

_upgrade = upgrade  # noqa: E305


@operation
def update(state=None, host=None):
    '''
    Updates apk repositories.
    '''

    yield 'apk update'

_update = update  # noqa: E305


@operation
def packages(
    packages=None, present=True, latest=False,
    update=False, upgrade=False,
    state=None, host=None,
):
    '''
    Add/remove/update apk packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``apk update`` before installing packages
    + upgrade: run ``apk upgrade`` before installing packages

    Versions:
        Package versions can be pinned like apk: ``<pkg>=<version>``.

    Examples:

    .. code:: python

        # Update package list and install packages
        apk.packages(
            name='Install Asterisk and Vim',
            packages=['asterisk', 'vim'],
            update=True,
        )

        # Install the latest versions of packages (always check)
        apk.packages(
            name='Install latest Vim',
            packages=['vim'],
            latest=True,
        )
    '''

    if update:
        yield _update(state=state, host=host)

    if upgrade:
        yield _upgrade(state=state, host=host)

    yield ensure_packages(
        host, packages, host.fact.apk_packages, present,
        install_command='apk add',
        uninstall_command='apk del',
        upgrade_command='apk upgrade',
        version_join='=',
        latest=latest,
    )



from __future__ import unicode_literals

import six


# NOTE https://github.com/python/cpython/blob/master/Lib/os.py#L1025
def fspath(path):
    path_types = six.string_types
    path_type_names = [pt.__name__ for pt in path_types]

    if isinstance(path, path_types):
        return path

    path_type = type(path)
    try:
        path_repr = path_type.__fspath__(path)
    except AttributeError:
        if hasattr(path_type, '__fspath__'):
            raise
        else:
            raise TypeError('expected {} or os.PathLike object, '
                            'not {}'.format(','.join(path_type_names), path_type.__name__))

    if isinstance(path_repr, path_types):
        return path_repr
    else:
        raise TypeError('expected {}.__fspath__() to return {} '
                        'not {}'.format(path_type.__name__,
                                        ','.join(path_type_names),
                                        type(path_repr).__name__))



from __future__ import division, print_function

import logging
import sys
import warnings

from fnmatch import fnmatch
from os import getcwd, path

import click

from pyinfra import (
    __version__,
    logger,
    pseudo_inventory,
    pseudo_state,
)
from pyinfra.api import State
from pyinfra.api.connect import connect_all, disconnect_all
from pyinfra.api.exceptions import NoGroupError, PyinfraError
from pyinfra.api.facts import (
    get_fact_class,
    get_fact_names,
    get_facts,
    is_fact,
    ShortFactBase,
)
from pyinfra.api.operation import add_op
from pyinfra.api.operations import run_ops
from pyinfra.operations import server

from .config import load_config, load_deploy_config
from .exceptions import (
    CliError,
    UnexpectedExternalError,
    UnexpectedInternalError,
)
from .inventory import make_inventory
from .log import setup_logging
from .prints import (
    print_facts,
    print_facts_list,
    print_inventory,
    print_meta,
    print_operations_list,
    print_results,
    print_state_facts,
    print_state_operations,
    print_support_info,
)
from .util import (
    get_operation_and_args,
    load_deploy_file,
)
from .virtualenv import init_virtualenv


# Exit handler
def _exit():
    if pseudo_state.isset() and pseudo_state.failed_hosts:
        sys.exit(1)
    sys.exit(0)


def _print_facts(ctx, param, value):
    if not value:
        return

    click.echo('--> Available facts:', err=True)
    print_facts_list()
    ctx.exit()


def _print_operations(ctx, param, value):
    if not value:
        return

    click.echo('--> Available operations:', err=True)
    print_operations_list()
    ctx.exit()


def _print_support(ctx, param, value):
    if not value:
        return

    click.echo('--> Support information:', err=True)
    print_support_info()
    ctx.exit()


@click.command()
@click.argument('inventory', nargs=1)
@click.argument('operations', nargs=-1, required=True)
@click.option(
    'verbosity', '-v',
    count=True,
    help='Print meta (-v), input (-vv) and output (-vvv).',
)
@click.option(
    '--dry', is_flag=True, default=False,
    help="Don't execute operations on the target hosts.",
)
@click.option(
    '--limit',
    help='Restrict the target hosts by name and group name.',
    multiple=True,
)
@click.option('--fail-percent', type=int, help='% of hosts allowed to fail.')
# Auth args
@click.option(
    '--sudo', is_flag=True, default=False,
    help='Whether to execute operations with sudo.',
)
@click.option('--sudo-user', help='Which user to sudo when sudoing.')
@click.option(
    '--use-sudo-password', is_flag=True, default=False,
    help='Whether to use a password with sudo.',
)
@click.option('--su-user', help='Which user to su to.')
@click.option('--shell-executable', help='Shell to use (ex: "sh", "cmd", "ps").')
# Operation flow args
@click.option('--parallel', type=int, help='Number of operations to run in parallel.')
@click.option(
    '--no-wait', is_flag=True, default=False,
    help="Don't wait between operations for hosts.",
)
@click.option(
    '--serial', is_flag=True, default=False,
    help='Run operations in serial, host by host.',
)
# SSH args
@click.option('--user', help='SSH user to connect as.')
@click.option('--port', type=int, help='SSH port to connect to.')
@click.option('--key', type=click.Path(), help='SSH Private key filename.')
@click.option('--key-password', help='SSH Private key password.')
@click.option('--password', help='SSH password.')
# WinRM args
@click.option('--winrm-username', help='WINRM user to connect as.')
@click.option('--winrm-password', help='WINRM password.')
@click.option('--winrm-port', help='WINRM port to connect to.')
# Eager commands (pyinfra [--facts | --operations | --support | --version])
@click.option(
    '--facts', is_flag=True, is_eager=True, callback=_print_facts,
    help='Print available facts list and exit.',
)
@click.option(
    'print_operations', '--operations',
    is_flag=True,
    is_eager=True,
    callback=_print_operations,
    help='Print available operations list and exit.',
)
@click.option(
    '--support', is_flag=True, is_eager=True, callback=_print_support,
    help='Print useful information for support and exit.',
)
# Debug args
@click.option(
    '--quiet', is_flag=True, default=False,
    help='Hide most pyinfra output.',
)
@click.option(
    '--debug', is_flag=True, default=False,
    help='Print debug info.',
)
@click.option(
    '--debug-data', is_flag=True, default=False,
    help='Print host/group data before connecting and exit.',
)
@click.option(
    '--debug-facts', is_flag=True, default=False,
    help='Print facts after generating operations and exit.',
)
@click.option(
    '--debug-operations', is_flag=True, default=False,
    help='Print operations after generating and exit.',
)
@click.version_option(
    version=__version__,
    prog_name='pyinfra',
    message='%(prog)s: v%(version)s',
)
def cli(*args, **kwargs):
    '''
    pyinfra manages the state of one or more servers. It can be used for
    app/service deployment, config management and ad-hoc command execution.

    Documentation: pyinfra.readthedocs.io

    # INVENTORY

    \b
    + a file (inventory.py)
    + hostname (host.net)
    + Comma separated hostnames:
      host-1.net,host-2.net,@local

    # OPERATIONS

    \b
    # Run one or more deploys against the inventory
    pyinfra INVENTORY deploy_web.py [deploy_db.py]...

    \b
    # Run a single operation against the inventory
    pyinfra INVENTORY server.user pyinfra home=/home/pyinfra

    \b
    # Execute an arbitrary command on the inventory
    pyinfra INVENTORY exec -- echo "hello world"

    \b
    # Run one or more facts on the inventory
    pyinfra INVENTORY fact linux_distribution [users]...
    pyinfra INVENTORY all-facts
    '''

    try:
        _main(*args, **kwargs)

    except PyinfraError as e:
        # Re-raise any internal exceptions that aren't handled by click as
        # CliErrors which are.
        if not isinstance(e, click.ClickException):
            message = getattr(e, 'message', e.args[0])
            raise CliError(message)

        raise

    except UnexpectedExternalError:
        # Pass unexpected external exceptions through as-is
        raise

    except Exception as e:
        # Re-raise any unexpected internal exceptions as UnexpectedInternalError
        raise UnexpectedInternalError(e)

    finally:
        if pseudo_state.isset() and pseudo_state.initialised:
            # Triggers any executor disconnect requirements
            disconnect_all(pseudo_state)


def _main(
    inventory, operations, verbosity,
    user, port, key, key_password, password,
    winrm_username, winrm_password, winrm_port,
    shell_executable,
    sudo, sudo_user, use_sudo_password, su_user,
    parallel, fail_percent,
    dry, limit, no_wait, serial, quiet,
    debug, debug_data, debug_facts, debug_operations,
    facts=None, print_operations=None, support=None,
):
    if not debug and not sys.warnoptions:
        warnings.simplefilter('ignore')

    # Setup logging
    log_level = logging.INFO
    if debug:
        log_level = logging.DEBUG
    elif quiet:
        log_level = logging.WARNING

    setup_logging(log_level)

    # Bootstrap any virtualenv
    init_virtualenv()

    deploy_dir = getcwd()
    potential_deploy_dirs = []

    # This is the most common case: we have a deploy file so use it's
    # pathname - we only look at the first file as we can't have multiple
    # deploy directories.
    if operations[0].endswith('.py'):
        deploy_file_dir, _ = path.split(operations[0])
        above_deploy_file_dir, _ = path.split(deploy_file_dir)

        deploy_dir = deploy_file_dir

        potential_deploy_dirs.extend((
            deploy_file_dir, above_deploy_file_dir,
        ))

    # If we have a valid inventory, look in it's path and it's parent for
    # group_data or config.py to indicate deploy_dir (--fact, --run).
    if inventory.endswith('.py') and path.isfile(inventory):
        inventory_dir, _ = path.split(inventory)
        above_inventory_dir, _ = path.split(inventory_dir)

        potential_deploy_dirs.extend((
            inventory_dir, above_inventory_dir,
        ))

    for potential_deploy_dir in potential_deploy_dirs:
        logger.debug('Checking potential directory: {0}'.format(
            potential_deploy_dir,
        ))

        if any((
            path.isdir(path.join(potential_deploy_dir, 'group_data')),
            path.isfile(path.join(potential_deploy_dir, 'config.py')),
        )):
            logger.debug('Setting directory to: {0}'.format(potential_deploy_dir))
            deploy_dir = potential_deploy_dir
            break

    # Make sure imported files (deploy.py/etc) behave as if imported from the cwd
    sys.path.append(deploy_dir)

    # Create an empty/unitialised state object
    state = State()
    # Set the deploy directory
    state.deploy_dir = deploy_dir

    pseudo_state.set(state)

    if verbosity > 0:
        state.print_fact_info = True
        state.print_noop_info = True

    if verbosity > 1:
        state.print_input = state.print_fact_input = True

    if verbosity > 2:
        state.print_output = state.print_fact_output = True

    if not quiet:
        click.echo('--> Loading config...', err=True)

    # Load up any config.py from the filesystem
    config = load_config(deploy_dir)

    # Make a copy before we overwrite
    original_operations = operations

    # Debug (print) inventory + group data
    if operations[0] == 'debug-inventory':
        command = 'debug-inventory'

    # Get all non-arg facts
    elif operations[0] == 'all-facts':
        command = 'fact'
        fact_names = []

        for fact_name in get_fact_names():
            fact_class = get_fact_class(fact_name)
            if (
                not issubclass(fact_class, ShortFactBase)
                and not callable(fact_class.command)
            ):
                fact_names.append(fact_name)

        operations = [(name, None) for name in fact_names]

    # Get one or more facts
    elif operations[0] == 'fact':
        command = 'fact'

        fact_names = operations[1:]
        facts = []

        for name in fact_names:
            args = None

            if ':' in name:
                name, args = name.split(':', 1)
                args = args.split(',')

            if not is_fact(name):
                raise CliError('No fact: {0}'.format(name))

            facts.append((name, args))

        operations = facts

    # Execute a raw command with server.shell
    elif operations[0] == 'exec':
        command = 'exec'
        operations = operations[1:]

    # Execute one or more deploy files
    elif all(cmd.endswith('.py') for cmd in operations):
        command = 'deploy'
        operations = operations[0:]

        for file in operations:
            if not path.exists(file):
                raise CliError('No deploy file: {0}'.format(file))

    # Operation w/optional args (<module>.<op> ARG1 ARG2 ...)
    elif len(operations[0].split('.')) == 2:
        command = 'op'
        operations = get_operation_and_args(operations)

    else:
        raise CliError('''Invalid operations: {0}

    Operation usage:
    pyinfra INVENTORY deploy_web.py [deploy_db.py]...
    pyinfra INVENTORY server.user pyinfra home=/home/pyinfra
    pyinfra INVENTORY exec -- echo "hello world"
    pyinfra INVENTORY fact os [users]...'''.format(operations))

    # Load any hooks/config from the deploy file
    if command == 'deploy':
        load_deploy_config(operations[0], config)

    # Arg based config overrides
    if sudo:
        config.SUDO = True
        if sudo_user:
            config.SUDO_USER = sudo_user

    if use_sudo_password:
        config.USE_SUDO_PASSWORD = use_sudo_password

    if su_user:
        config.SU_USER = su_user

    if parallel:
        config.PARALLEL = parallel

    if shell_executable:
        config.SHELL = shell_executable

    if fail_percent is not None:
        config.FAIL_PERCENT = fail_percent

    if not quiet:
        click.echo('--> Loading inventory...', err=True)

    # Load up the inventory from the filesystem
    inventory, inventory_group = make_inventory(
        inventory,
        deploy_dir=deploy_dir,
        ssh_port=port,
        ssh_user=user,
        ssh_key=key,
        ssh_key_password=key_password,
        ssh_password=password,
        winrm_username=winrm_username,
        winrm_password=winrm_password,
        winrm_port=winrm_port,
    )

    # Attach to pseudo inventory
    pseudo_inventory.set(inventory)

    # Now that we have inventory, apply --limit config override
    initial_limit = None
    if limit:
        all_limit_hosts = []

        for limiter in limit:
            try:
                limit_hosts = inventory.get_group(limiter)
            except NoGroupError:
                limits = limiter.split(',')
                if len(limits) > 1:
                    logger.warning((
                        'Specifying comma separated --limit values is deprecated, '
                        'please use multiple --limit options.'
                    ))

                limit_hosts = [
                    host for host in inventory
                    if any(fnmatch(host.name, match) for match in limits)
                ]

            all_limit_hosts.extend(limit_hosts)

        initial_limit = list(set(all_limit_hosts))

    # Initialise the state
    state.init(inventory, config, initial_limit=initial_limit)

    # If --debug-data dump & exit
    if command == 'debug-inventory' or debug_data:
        if debug_data:
            logger.warning((
                '--debug-data is deprecated, '
                'please use `pyinfra INVENTORY debug-inventory` instead.'
            ))
        print_inventory(state)
        _exit()

    # Connect to all the servers
    if not quiet:
        click.echo(err=True)
        click.echo('--> Connecting to hosts...', err=True)
    connect_all(state)

    # Just getting a fact?
    #

    if command == 'fact':
        if not quiet:
            click.echo(err=True)
            click.echo('--> Gathering facts...', err=True)

        state.print_fact_info = True
        fact_data = {}

        for i, command in enumerate(operations):
            name, args = command
            fact_key = name
            if args:
                fact_key = '{0}{1}'.format(name, tuple(args))
            try:
                fact_data[fact_key] = get_facts(
                    state, name,
                    args=args,
                    apply_failed_hosts=False,
                )
            except PyinfraError:
                pass

        print_facts(fact_data)
        _exit()

    # Prepare the deploy!
    #

    # Execute a raw command with server.shell
    if command == 'exec':
        # Print the output of the command
        state.print_output = True

        add_op(
            state,
            server.shell,
            ' '.join(operations),
            _allow_cli_mode=True,
        )

    # Deploy files(s)
    elif command == 'deploy':
        if not quiet:
            click.echo(err=True)
            click.echo('--> Preparing operations...', err=True)

        # Number of "steps" to make = number of files * number of hosts
        for i, filename in enumerate(operations):
            logger.info('Loading: {0}'.format(click.style(filename, bold=True)))
            load_deploy_file(state, filename)

    # Operation w/optional args
    elif command == 'op':
        if not quiet:
            click.echo(err=True)
            click.echo('--> Preparing operation...', err=True)

        op, args = operations
        args, kwargs = args
        kwargs['_allow_cli_mode'] = True

        def print_host_ready(host):
            logger.info('{0}{1} {2}'.format(
                host.print_prefix,
                click.style('Ready:', 'green'),
                click.style(original_operations[0], bold=True),
            ))

        kwargs['_after_host_callback'] = print_host_ready

        add_op(state, op, *args, **kwargs)

    # Always show meta output
    if not quiet:
        click.echo(err=True)
        click.echo('--> Proposed changes:', err=True)
    print_meta(state)

    # If --debug-facts or --debug-operations, print and exit
    if debug_facts or debug_operations:
        if debug_facts:
            print_state_facts(state)

        if debug_operations:
            print_state_operations(state)

        _exit()

    # Run the operations we generated with the deploy file
    if dry:
        _exit()

    if not quiet:
        click.echo(err=True)

    if not quiet:
        click.echo('--> Beginning operation run...', err=True)
    run_ops(state, serial=serial, no_wait=no_wait)

    if not quiet:
        click.echo('--> Results:', err=True)
    print_results(state)

    _exit()



import pkg_resources

# Connectors that handle generation of inventories
ALL_CONNECTORS = {
    entrypoint.name: entrypoint.load()
    for entrypoint in pkg_resources.iter_entry_points('pyinfra.connectors')
}


# Connectors that handle execution of pyinfra operations
EXECUTION_CONNECTORS = {
    connector: connector_mod
    for connector, connector_mod in ALL_CONNECTORS.items()
    if getattr(connector_mod, 'EXECUTION_CONNECTOR', False)
}



from __future__ import print_function, unicode_literals

import traceback

from itertools import product
from socket import (
    error as socket_error,
    timeout as timeout_error,
)

import click
import gevent
import six

from paramiko import SSHException

import pyinfra

from pyinfra import logger
from pyinfra.progress import progress_spinner

from .command import (
    FunctionCommand,
    PyinfraCommand,
    StringCommand,
)
from .exceptions import PyinfraError
from .operation_kwargs import get_executor_kwarg_keys
from .util import (
    format_exception,
    log_error_or_warning,
    log_host_command_error,
    memoize,
    print_host_combined_output,
)


@memoize
def show_pre_or_post_condition_warning(condition_name):
    logger.warning('The `{0}` argument is in beta!'.format(condition_name))


def _run_shell_command(state, host, command, global_kwargs, executor_kwargs):
    status = False
    combined_output_lines = []

    try:
        status, combined_output_lines = command.execute(state, host, executor_kwargs)
    except (timeout_error, socket_error, SSHException) as e:
        log_host_command_error(
            host,
            e,
            timeout=global_kwargs['timeout'],
        )

    # If we failed and have no already printed the stderr, print it
    if status is False and not state.print_output:
        print_host_combined_output(host, combined_output_lines)

    return status


def _run_server_op(state, host, op_hash):
    state.trigger_callbacks('operation_host_start', host, op_hash)

    if op_hash not in state.ops[host]:
        logger.info('{0}{1}'.format(host.print_prefix, click.style('Skipped', 'blue')))
        return True

    op_data = state.get_op_data(host, op_hash)
    global_kwargs = op_data['global_kwargs']

    op_meta = state.get_op_meta(op_hash)

    ignore_errors = global_kwargs['ignore_errors']

    logger.debug('Starting operation {0} on {1}'.format(
        ', '.join(op_meta['names']), host,
    ))

    executor_kwarg_keys = get_executor_kwarg_keys()
    base_executor_kwargs = {
        key: global_kwargs[key]
        for key in executor_kwarg_keys
        if key in global_kwargs
    }

    precondition = global_kwargs['precondition']
    if precondition:
        show_pre_or_post_condition_warning('precondition')
    if precondition and not _run_shell_command(
        state, host, StringCommand(precondition), global_kwargs, base_executor_kwargs,
    ):
        log_error_or_warning(
            host, ignore_errors,
            description='precondition failed: {0}'.format(precondition),
        )
        if not ignore_errors:
            state.trigger_callbacks('operation_host_error', host, op_hash)
            return False

    state.ops_run.add(op_hash)

    # ...loop through each command
    for i, command in enumerate(op_data['commands']):

        status = False

        executor_kwargs = base_executor_kwargs.copy()
        executor_kwargs.update(command.executor_kwargs)

        # Now we attempt to execute the command
        #

        if not isinstance(command, PyinfraCommand):
            raise TypeError('{0} is an invalid pyinfra command!'.format(command))

        if isinstance(command, FunctionCommand):
            try:
                status = command.execute(state, host, executor_kwargs)
            except Exception as e:  # Custom functions could do anything, so expect anything!
                logger.warning(traceback.format_exc())
                logger.error('{0}{1}'.format(
                    host.print_prefix,
                    click.style(
                        'Unexpected error in Python callback: {0}'.format(
                            format_exception(e),
                        ),
                        'red',
                    ),
                ))

        elif isinstance(command, StringCommand):
            status = _run_shell_command(state, host, command, global_kwargs, executor_kwargs)

        else:
            try:
                status = command.execute(state, host, executor_kwargs)
            except (timeout_error, socket_error, SSHException, IOError) as e:
                log_host_command_error(
                    host,
                    e,
                    timeout=global_kwargs['timeout'],
                )

        # Break the loop to trigger a failure
        if status is False:
            break

        state.results[host]['commands'] += 1

    # Commands didn't break, so count our successes & return True!
    else:
        postcondition = global_kwargs['postcondition']
        if postcondition:
            show_pre_or_post_condition_warning('postcondition')
        if postcondition and not _run_shell_command(
            state, host, StringCommand(postcondition), global_kwargs, base_executor_kwargs,
        ):
            log_error_or_warning(
                host, ignore_errors,
                description='postcondition failed: {0}'.format(postcondition),
            )
            if not ignore_errors:
                state.trigger_callbacks('operation_host_error', host, op_hash)
                return False

        # Count success
        state.results[host]['ops'] += 1
        state.results[host]['success_ops'] += 1

        logger.info('{0}{1}'.format(
            host.print_prefix,
            click.style(
                'Success' if len(op_data['commands']) > 0 else 'No changes',
                'green',
            ),
        ))

        # Trigger any success handler
        if global_kwargs['on_success']:
            global_kwargs['on_success'](state, host, op_hash)

        state.trigger_callbacks('operation_host_success', host, op_hash)
        return True

    # Up error_ops & log
    state.results[host]['error_ops'] += 1

    log_error_or_warning(host, ignore_errors)

    # Always trigger any error handler
    if global_kwargs['on_error']:
        global_kwargs['on_error'](state, host, op_hash)

    # Ignored, op "completes" w/ ignored error
    if ignore_errors:
        state.results[host]['ops'] += 1

    # Unignored error -> False
    state.trigger_callbacks('operation_host_error', host, op_hash)
    if ignore_errors:
        return True
    return False


def _log_operation_start(op_meta):
    op_types = []
    if op_meta['serial']:
        op_types.append('serial')
    if op_meta['run_once']:
        op_types.append('run once')

    args = ''
    if op_meta['args']:
        args = '({0})'.format(', '.join(str(arg) for arg in op_meta['args']))

    logger.info('{0} {1} {2}'.format(
        click.style('--> Starting{0}operation:'.format(
            ' {0} '.format(', '.join(op_types)) if op_types else ' ',
        ), 'blue'),
        click.style(', '.join(op_meta['names']), bold=True),
        args,
    ))


def _run_server_ops(state, host, progress=None):
    '''
    Run all ops for a single server.
    '''

    logger.debug('Running all ops on {0}'.format(host))

    for op_hash in state.get_op_order():
        op_meta = state.get_op_meta(op_hash)
        _log_operation_start(op_meta)

        result = _run_server_op(state, host, op_hash)

        # Trigger CLI progress if provided
        if progress:
            progress((host, op_hash))

        if result is False:
            raise PyinfraError('Error in operation {0} on {1}'.format(
                ', '.join(op_meta['names']), host,
            ))

        if pyinfra.is_cli:
            click.echo(err=True)


def _run_serial_ops(state):
    '''
    Run all ops for all servers, one server at a time.
    '''

    for host in list(state.inventory.iter_active_hosts()):
        host_operations = product([host], state.get_op_order())
        with progress_spinner(host_operations) as progress:
            try:
                _run_server_ops(
                    state, host,
                    progress=progress,
                )
            except PyinfraError:
                state.fail_hosts({host})


def _run_no_wait_ops(state):
    '''
    Run all ops for all servers at once.
    '''

    hosts_operations = product(state.inventory.iter_active_hosts(), state.get_op_order())
    with progress_spinner(hosts_operations) as progress:
        # Spawn greenlet for each host to run *all* ops
        greenlets = [
            state.pool.spawn(
                _run_server_ops, state, host,
                progress=progress,
            )
            for host in state.inventory.iter_active_hosts()
        ]
        gevent.joinall(greenlets)


def _run_single_op(state, op_hash):
    '''
    Run a single operation for all servers. Can be configured to run in serial.
    '''

    state.trigger_callbacks('operation_start', op_hash)

    op_meta = state.get_op_meta(op_hash)
    _log_operation_start(op_meta)

    failed_hosts = set()

    if op_meta['serial']:
        with progress_spinner(state.inventory.iter_active_hosts()) as progress:
            # For each host, run the op
            for host in state.inventory.iter_active_hosts():
                result = _run_server_op(state, host, op_hash)
                progress(host)

                if not result:
                    failed_hosts.add(host)

    else:
        # Start with the whole inventory in one batch
        batches = [list(state.inventory.iter_active_hosts())]

        # If parallel set break up the inventory into a series of batches
        if op_meta['parallel']:
            parallel = op_meta['parallel']
            hosts = list(state.inventory.iter_active_hosts())

            batches = [
                hosts[i:i + parallel]
                for i in range(0, len(hosts), parallel)
            ]

        for batch in batches:
            with progress_spinner(batch) as progress:
                # Spawn greenlet for each host
                greenlet_to_host = {
                    state.pool.spawn(_run_server_op, state, host, op_hash): host
                    for host in batch
                }

                # Trigger CLI progress as hosts complete if provided
                for greenlet in gevent.iwait(greenlet_to_host.keys()):
                    host = greenlet_to_host[greenlet]
                    progress(host)

                # Get all the results
                for greenlet, host in six.iteritems(greenlet_to_host):
                    if not greenlet.get():
                        failed_hosts.add(host)

    # Now all the batches/hosts are complete, fail any failures
    state.fail_hosts(failed_hosts)

    if pyinfra.is_cli:
        click.echo(err=True)

    state.trigger_callbacks('operation_end', op_hash)


def run_ops(state, serial=False, no_wait=False):
    '''
    Runs all operations across all servers in a configurable manner.

    Args:
        state (``pyinfra.api.State`` obj): the deploy state to execute
        serial (boolean): whether to run operations host by host
        no_wait (boolean): whether to wait for all hosts between operations
    '''

    # Flag state as deploy in process
    state.deploying = True

    # Run all ops, but server by server
    if serial:
        _run_serial_ops(state)

    # Run all the ops on each server in parallel (not waiting at each operation)
    elif no_wait:
        _run_no_wait_ops(state)

    # Default: run all ops in order, waiting at each for all servers to complete
    else:
        for op_hash in state.get_op_order():
            _run_single_op(state, op_hash)



from os import path
from unittest import TestCase

from pyinfra.api.connectors.ansible import make_names_data
from pyinfra.api.exceptions import InventoryError

TEST_DATA = [
    ('mail.example.com', {}, ['all']),
    ('foo.example.com', {}, ['all', 'webservers']),
    ('bar.example.com', {}, ['all', 'webservers']),
    ('one.example.com', {}, ['all', 'dbservers']),
    ('two.example.com', {}, ['all', 'dbservers']),
    ('three.example.com', {}, ['all', 'dbservers']),
]


class TestAnsibleConnector(TestCase):
    def test_make_names_data_no_file(self):
        with self.assertRaises(InventoryError):
            make_names_data(inventory_filename='/not/a/file')

    def test_make_names_data_ini(self):
        data = make_names_data(inventory_filename=path.join(
            'tests', 'deploy', 'inventories', 'inventory_ansible',
        ))

        test_data = [
            ('webserver-1.net', {}, ['web_and_db_servers', 'webservers']),
            ('webserver-2.net', {}, ['web_and_db_servers', 'webservers']),
            ('webserver-3.net', {}, ['web_and_db_servers', 'webservers']),
            ('dbserver-01.net', {}, ['dbservers', 'web_and_db_servers']),
            ('dbserver-02.net', {}, ['dbservers', 'web_and_db_servers']),
        ]

        assert sorted(data) == sorted(test_data)

    def test_make_names_data_json(self):
        data = make_names_data(inventory_filename=path.join(
            'tests', 'deploy', 'inventories', 'inventory_ansible.json',
        ))

        assert sorted(data) == sorted(TEST_DATA)

    def test_make_names_data_yaml(self):
        data = make_names_data(inventory_filename=path.join(
            'tests', 'deploy', 'inventories', 'inventory_ansible.yml',
        ))

        assert sorted(data) == sorted(TEST_DATA)



'''
The Python module allows you to execute Python code within the context of a deploy.
'''

from pyinfra.api import FunctionCommand, operation


@operation
def call(function, *args, **kwargs):
    '''
    Execute a Python function within a deploy.

    + function: the function to execute
    + args: additional arguments to pass to the function
    + kwargs: additional keyword arguments to pass to the function

    Callback functions args passed the state, host, and any args/kwargs passed
    into the operation directly, eg:

    .. code:: python

        def my_callback(state, host, hello=None):
            command = 'echo hello'
            if hello:
                command = command + ' ' + hello

            status, stdout, stderr = host.run_shell_command(command=command, sudo=SUDO)
            assert status is True  # ensure the command executed OK

            if 'hello ' not in '\\n'.join(stdout):  # stdout/stderr is a *list* of lines
                raise Exception(
                    f'`{command}` problem with callback stdout:{stdout} stderr:{stderr}',
                )

        python.call(
            name='Run my_callback function',
            function=my_callback,
            hello='world',
        )

    '''

    kwargs.pop('state', None)
    kwargs.pop('host', None)
    yield FunctionCommand(function, args, kwargs)


@operation
def raise_exception(exception, *args, **kwargs):
    def raise_exc(*args, **kwargs):  # pragma: no cover
        raise exception(*args, **kwargs)

    kwargs.pop('state', None)
    kwargs.pop('host', None)
    yield FunctionCommand(raise_exc, args, kwargs)



'''
The windows module handles misc windows operations.
'''

from __future__ import unicode_literals

from pyinfra.api import operation

# Tip: Use 'Get-Command -Noun Service' to search for what commands are available or
# simply 'Get-Command' to see what you can do...)

# Tip: To see the windows help page about a command, use 'Get-Help'.
# Might have to run 'Update-Help' if you want to use arguments like '-Examples'.
# ex: 'Get-Help Stop-Service'
# ex: 'Get-Help Stop-Service -Examples'
# ex: 'Get-Help Stop-Service -Detailed'
# ex: 'Get-Help Stop-Service -Full'

# FUTURE: add ability to stop processes (ex: "Stop-Process <id>")


@operation
def service(name, running=True, restart=False, suspend=False, state=None, host=None):
    '''
    Stop/Start a Windows service.

    + name: name of the service to manage
    + running: whether the the service should be running or stopped
    + restart: whether the the service should be restarted
    + suspend: whether the the service should be suspended

    Example:

    .. code:: python

        windows.service(
            {'Stop the spooler service'},
            'service',
            running=False,
        )
    '''

    if suspend or not running:
        if suspend:
            yield 'Suspend-Service -Name {0}'.format(name)
        else:
            yield 'Stop-Service -Name {0}'.format(name)
    else:
        if restart:
            yield 'Restart-Service -Name {0}'.format(name)
        else:
            if running:
                yield 'Start-Service -Name {0}'.format(name)


@operation
def reboot(state=None, host=None):
    '''
    Restart the server.
    '''
    yield 'Restart-Computer -Force'



from pyinfra.api import FactBase

from .util.packaging import parse_packages

CHOCO_REGEX = r'^([a-zA-Z0-9\.\-\+\_]+)\s([0-9\.]+)$'


class ChocoPackages(FactBase):
    '''
    Returns a dict of installed choco (Chocolatey) packages:

    .. code:: python

        {
            'package_name': ['version'],
        }
    '''

    command = 'choco list --local-only'
    shell_executable = 'ps'

    default = dict

    def process(self, output):
        return parse_packages(CHOCO_REGEX, output)


class ChocoVersion(FactBase):
    '''
    Returns the choco (Chocolatey) version.
    '''

    command = 'choco --version'

    @staticmethod
    def process(output):
        return ''.join(output).replace('\n', '')
